import platform
import re
from enum import Enum, auto
from hashlib import md5
from operator import eq, ge, gt, le, lt, ne
from pathlib import Path
from typing import Annotated, Any, Callable, List, Optional, Union, get_args, get_origin

from pydantic import ValidationInfo
from pydantic.fields import FieldInfo
from strenum import StrEnum

SCIENTIFIC_NOTATION_PATTERN = r"([\d.]+)([dD])([+-]?\d{1,3})"

# matches a float: 1d9, 1D-3, 1.D+4, etc.
SCIENTIFIC_NOTATION_REGEX = re.compile(SCIENTIFIC_NOTATION_PATTERN)

PYTHON_STYLES = r"\1e\3"

valid_types = (
    float,
    list[float],
    List[float],
    Optional[float],
    Optional[List[float]],
    Optional[list[float]],
)


def to_key(string: str) -> str:
    """
    Construct a key name from a given field name.
    The given field name may be a Pydantic Field alias, and the key name
    is intended to be used as a BaseModel class member variable name.

    Args:
        string (str): input field name

    """
    # First replace any leading digits, because those are undesirable
    # in variable names.
    digitdict = {
        "0": "zero",
        "1": "one",
        "2": "two",
        "3": "three",
        "4": "four",
        "5": "five",
        "6": "six",
        "7": "seven",
        "8": "eight",
        "9": "nine",
    }
    m = re.search(r"^\d+", string)
    if m:
        digitstring = string[0 : m.end()]  # noqa: E203
        for key, val in digitdict.items():
            digitstring = digitstring.replace(key, val)
        string = digitstring + string[m.end() :]  # noqa: E203

    # Next, replace spaces and hyphens in the potential variable name.
    return string.lower().replace(" ", "_").replace("-", "")


def to_list(item: Any) -> List[Any]:
    """Puts the specified item in a list if it is an instance of `dict`.

    Attributes:
        item: The item to put in a list.

    Returns:
        List: A list with the specified item.
    """

    if not isinstance(item, list):
        return [item]
    return item


def str_is_empty_or_none(str_field: str) -> bool:
    """
    Verifies whether a string is empty or None.

    Args:
        str_field (str): String to validate.

    Returns:
        bool: Evaluation result.
    """
    return str_field is None or not str_field or str_field.isspace()


def get_str_len(str_field: Optional[str]) -> int:
    """
    Get string length or 0 if input is None.

    Args:
        str_field (str): String to measure.

    Returns:
        int: Length of passed input.
    """
    return len(str_field) if str_field else 0


def get_substring_between(source: str, start: str, end: str) -> Optional[str]:
    """Finds the substring between two other strings.

    Args:
        source (str): The source string.
        start (str): The starting string from where to create the substring.
        end (str): The end string to where to create the substring.

    Returns:
        str: The substring if found; otherwise, `None`.
    """

    index_start = source.find(start)
    if index_start == -1:
        return None

    index_start += len(start)

    index_end = source.rfind(end, index_start)
    if index_end == -1:
        return None

    return source[index_start:index_end]


def operator_str(operator_func: Callable) -> str:
    """
    Make string representation of some of operator's built-in operator
    functions, for use in prettyprinting.

    Args:
        operator_func (Callable): Typically one of operator's built-in
            operator functions. When unsupported, the standard __str__
            representation is returned.
    """
    if operator_func == eq:
        return "is"
    elif operator_func == ne:
        return "is not"
    elif operator_func == lt:
        return "is less than"
    elif operator_func == le:
        return "is less than or equal to"
    elif operator_func == gt:
        return "is greater than"
    elif operator_func == ge:
        return "is greater than or equal to"
    else:
        return str(operator_func)


def resolve_file_model(
    value: Union[str, Path], model: Callable[[Union[str, Path]], Any]
):
    """Resolve file model.

     Resolves a file model based on the provided value and context.
     This function determines whether to use a `DiskOnlyFileModel` or the
     provided `model` class to resolve the given file path or string, depending
     on the current file load context settings.

     Args:
         value (Union[str, Path]):
            The file path or string to be resolved.
         model:
            The model class to use for resolving the file if the context settings allow recursion.
     Returns:
         An instance of `DiskOnlyFileModel` or the provided `model` class, depending on the file load context settings.

    Notes:
        - The function choose the `DiskOnlyFileModel` when the context's load settings do not allow recursion.
    """
    from hydrolib.core.base.file_manager import file_load_context
    from hydrolib.core.base.models import DiskOnlyFileModel

    with file_load_context() as context:
        if (
            hasattr(context, "_load_settings")
            and context._load_settings is not None
            and not context._load_settings.recurse
        ):
            result = DiskOnlyFileModel(value)
        else:
            result = model(value)
    return result


class PathToDictionaryConverter:

    @staticmethod
    def convert(cls, value: Any, info: ValidationInfo):
        """Convert a value to a dictionary if it is a file model type.

        Args:
            cls (Type):
                The class to which the value belongs.
            value (Any):
                The value to convert.
            info (ValidationInfo):
                Validation information.

        Returns:
            Any:
                The converted value, which is a dictionary if the value is a file model type.
        """
        from hydrolib.core.dflowfm.ini.util import split_string_on_delimiter

        fields = cls.model_fields
        key = info.field_name

        if isinstance(value, (str, Path, list)) and fields.get(key) is not None:
            if PathToDictionaryConverter.is_file_model_type(fields[key].annotation):
                value = PathToDictionaryConverter.make_dict(value)
            elif PathToDictionaryConverter.is_list_file_model_type(
                fields[key].annotation
            ):
                value = [
                    (
                        PathToDictionaryConverter.make_dict(v)
                        if isinstance(v, (str, Path))
                        else v
                    )
                    for v in split_string_on_delimiter(cls, value, info)
                ]

        return value

    @staticmethod
    def make_dict(value: Union[str, Path]) -> Union[dict, "DiskOnlyFileModel"]:
        """Convert a value to a dictionary with a 'filepath' key.

        Args:
            value (Union[str, Path]): The value to convert, which can be a string or a Path object.

        Returns:
            dict: A dictionary with a 'filepath' key containing the Path object.
            DiskOnlyFileModel: If the context's load settings do not recurse, return a DiskOnlyFileModel.
        """
        from hydrolib.core.base.file_manager import file_load_context
        from hydrolib.core.base.models import DiskOnlyFileModel

        with file_load_context() as context:
            if (
                hasattr(context, "_load_settings")
                and context._load_settings is not None
                and not context._load_settings.recurse
            ):
                value = DiskOnlyFileModel(value)
            else:
                value = {"filepath": Path(value)}
        return value

    @staticmethod
    def is_file_model_type(annotation: Any) -> bool:
        """Check if the given annotation is a FileModel type.

        Args:
            annotation (Any): The annotation to check.

        Returns:
            bool: True if the annotation is a FileModel type, False otherwise.
        """
        from hydrolib.core.base.models import FileModel

        stack = [annotation]
        result = False
        while stack:
            current = stack.pop()
            origin = get_origin(current)
            if origin is Union:
                stack.append(get_args(current)[0])
                continue
            if origin is Annotated:
                stack.append(get_args(current)[0])
                continue
            if isinstance(current, type) and issubclass(current, FileModel):
                result = True
        return result

    @staticmethod
    def is_list_file_model_type(annotation: Any) -> bool:
        """Check if the given annotation is a list of FileModel types.

        Args:
            annotation (Any): The annotation to check.

        Returns:
            bool: True if the annotation is a list of FileModel types, False otherwise.
        """
        stack = [annotation]
        result = False
        while stack:
            current = stack.pop()
            origin = get_origin(current)
            if origin is Union:
                stack.append(get_args(current)[0])
                continue
            if origin is Annotated:
                stack.append(get_args(current)[0])
                continue
            if origin is list:
                result = PathToDictionaryConverter.is_file_model_type(
                    get_args(current)[0]
                )
        return result


class OperatingSystem(Enum):
    """Enum represting an operating system."""

    WINDOWS = auto()
    """The Windows operating system."""
    LINUX = auto()
    """The Linux operating system."""
    MACOS = auto()
    """The MacOS operating system."""


def get_operating_system() -> OperatingSystem:
    """Gets the currently running operating system.

    Raises:
        NotImplementedError: When the current operating system is anything other than Windows, Linux or MacOS.

    Returns:
        OperatingSystem: The operating system.
    """
    operating_system = platform.system()

    if operating_system == "Windows":
        return OperatingSystem.WINDOWS
    if operating_system == "Linux":
        return OperatingSystem.LINUX
    if operating_system == "Darwin":
        return OperatingSystem.MACOS

    raise NotImplementedError(f"Operating system {operating_system} is not supported.")


class PathStyle(StrEnum):
    """Path style format."""

    UNIXLIKE = "unix"
    """Unix-like path style."""
    WINDOWSLIKE = "windows"
    """Windows-like path style."""


def get_path_style_for_current_operating_system() -> PathStyle:
    """Gets the file path style for the currently running operating system.

    Returns:
        OperatingSystem: The OS path style.
    """
    operating_system = get_operating_system()

    if operating_system == OperatingSystem.WINDOWS:
        return PathStyle.WINDOWSLIKE
    else:
        return PathStyle.UNIXLIKE


class FilePathStyleConverter:
    """Class for converting file paths between different path styles."""

    def __init__(self):
        """Initialize the converter with the current operating system's path style."""
        self._os_path_style = get_path_style_for_current_operating_system()

    def convert_to_os_style(self, file_path: Path, source_path_style: PathStyle) -> str:
        """Convert the file path from the source path style to the path style of the current operating system.

        Args:
            file_path (Path):
                The file path to convert to the OS path style.
            source_path_style (PathStyle):
                The file path style of the given file path.

        Returns:
            str: The converted file path with OS path style.

        Raises:
            NotImplementedError: When this function is called with a PathStyle other than WINDOWSLIKE or UNIXLIKE.
        """

        return FilePathStyleConverter._convert(
            file_path, source_path_style, self._os_path_style
        )

    def convert_from_os_style(
        self, file_path: Path, target_path_style: PathStyle
    ) -> str:
        """Convert the file path from the path style of the current operating system to the target path style.

        Args:
            file_path (Path): The file path to convert to the target path style.
            target_path_style (PathStyle): The target path style to convert the file path to.

        Returns:
            str: The converted file path with the target path style.

        Raises:
            NotImplementedError: When this function is called with a PathStyle other than WINDOWSLIKE or UNIXLIKE.
        """

        return FilePathStyleConverter._convert(
            file_path, self._os_path_style, target_path_style
        )

    @classmethod
    def _convert(
        cls, file_path: Path, source_path_style: PathStyle, target_path_style: PathStyle
    ) -> str:
        if source_path_style == target_path_style:
            return file_path.as_posix()

        if (
            source_path_style == PathStyle.UNIXLIKE
            and target_path_style == PathStyle.WINDOWSLIKE
        ):
            return FilePathStyleConverter._from_posix_to_windows_path(file_path)
        elif (
            source_path_style == PathStyle.WINDOWSLIKE
            and target_path_style == PathStyle.UNIXLIKE
        ):
            return FilePathStyleConverter._from_windows_to_posix_path(file_path)
        else:
            raise NotImplementedError(
                f"Cannot convert {source_path_style} to {target_path_style}"
            )

    @classmethod
    def _from_posix_to_windows_path(cls, posix_path: Path) -> str:
        is_relative = not posix_path.as_posix().startswith("/")

        if is_relative:
            return posix_path.as_posix()

        root = posix_path.parts[1]
        windows_root = root + ":/"
        parts = posix_path.parts[2:]

        windows_path = windows_root + "/".join(parts)

        return windows_path

    @classmethod
    def _from_windows_to_posix_path(cls, windows_path: Path) -> str:
        windows_path_str = str(windows_path).replace("\\", "/")
        windows_path = Path(windows_path_str)

        root = windows_path.parts[0]
        is_relative = ":" not in root

        if is_relative:
            return windows_path_str

        posix_root = "/" + root.split(":")[0] + "/"
        parts = windows_path.parts[1:]

        posix_path = posix_root + "/".join(parts)

        return posix_path


class FileChecksumCalculator:
    """
    FileChecksumCalculator calculator used to calculate the checksum of a file.
    """

    @staticmethod
    def calculate_checksum(filepath: Path) -> Optional[str]:
        """Calculate the checksum of the file from the given filepath.

        Args:
            filepath (Path): The filepath to the file for which the checksum will be calculated.

        Returns:
            [Optional[str]]:
                The checksum of the file.
                When the filepath doesn't exist or the filepath isn't a file, None.
        """
        if not filepath.exists() or not filepath.is_file():
            return None

        return FileChecksumCalculator._calculate_md5_checksum(filepath)

    @staticmethod
    def _calculate_md5_checksum(filepath: Path) -> str:
        """Calculate the MD5 checksum of a file.

        This method uses the `hashlib.md5` function with the `usedforsecurity=False` parameter
        to indicate that the MD5 hash is not being used for security purposes. MD5 is considered
        cryptographically insecure and should not be used in security-sensitive contexts.

        Note:
            The `usedforsecurity` parameter was introduced in Python 3.9. This method requires
            Python 3.9 or later to function correctly.

        Args:
            filepath (Path): The path to the file for which the checksum will be calculated.

        Returns:
            str: The MD5 checksum of the file.
        """
        md5_hash = md5(usedforsecurity=False)
        with open(filepath, "rb") as file:
            for chunk in iter(lambda: file.read(4096), b""):
                md5_hash.update(chunk)
        return md5_hash.hexdigest()


class FortranUtils:
    """Utility class for Fortran specific conventions."""

    @staticmethod
    def replace_fortran_scientific_notation(value):
        """Replace Fortran scientific notation ("D" in exponent) with standard
        scientific notation ("e" in exponent).
        """
        if isinstance(value, str):
            return SCIENTIFIC_NOTATION_REGEX.sub(PYTHON_STYLES, value)
        elif isinstance(value, list):
            return list(map(FortranUtils.replace_fortran_scientific_notation, value))

        return value


class FortranScientificNotationConverter:

    @classmethod
    def convert(cls, value: Union[str, List[str]]) -> Union[str, List[str]]:
        """
        Replaces FORTRAN-style scientific notation in a value.

        Args:
            value (Any): The value to process.

        Returns:
            Any: The processed value.
        """
        if isinstance(value, str):
            return SCIENTIFIC_NOTATION_REGEX.sub(PYTHON_STYLES, value)
        if isinstance(value, list):
            for i, v in enumerate(value):
                if isinstance(v, str):
                    value[i] = SCIENTIFIC_NOTATION_REGEX.sub(PYTHON_STYLES, v)

        return value

    @classmethod
    def convert_fields(cls, values: dict, field_definitions: dict) -> dict:
        """Convert Fields

        Converts values in a dictionary using Fortran-style scientific notation
        to Python floats for fields defined as float or List[float].

        Args:
            values (dict):
                The input dictionary of field values.
            field_definitions (dict):
                A mapping of field names to Pydantic Field definitions.

        Returns:
            dict:
                The updated dictionary with converted float fields.
        """
        alias_to_field = {
            field.alias: name
            for name, field in field_definitions.items()
            if field.alias is not None
        }
        new_values = {}
        for field_name, value in values.items():
            actual_field_name = alias_to_field.get(field_name, field_name)
            field: FieldInfo = field_definitions.get(actual_field_name)
            if field:
                field_type = field.annotation
                if field_type not in valid_types:
                    new_values[field_name] = value
                else:
                    # convert only the value if it is a float or a list of floats
                    new_values[field_name] = cls.convert(value)
            else:
                new_values[field_name] = value
        return new_values



from pathlib import Path

from hydrolib.core.base.models import ModelSaveSettings, SerializerConfig


class T3DSerializerConfig(SerializerConfig):
    """Configuration settings for the TimSerializer."""

    column_spacing: int = 1
    """(int): The number of spaces to include between columns in the serialized .t3d file."""


class T3DSerializer:

    @staticmethod
    def serialize(
        path: Path,
        data: dict,
        config: SerializerConfig,
        save_settings: ModelSaveSettings,
    ) -> None:
        """
        Serialize the T3D model data (comments, layer_type, layers, and records)
        into the file at `path`, conforming to the typical .t3d format.

        Args:
            path (Path): Where to write the .t3d file.
            data (dict): A dict representation of the T3DModel. Contains:
                {
                  "comments": [str, ...],
                  "layer_type": "SIGMA" or "Z",
                  "layers": [float, float, ...],
                  "vectormax": Optional[int],
                  "records": [
                    {
                      "time": str,
                      "data": [floats...]
                    }, ...
                  ]
                }
            config (SerializerConfig): Contains settings like float_format.
            save_settings (ModelSaveSettings): Additional settings controlling how paths are saved, etc.

        Examples:
            ```python
            >>> from pathlib import Path
            >>> from hydrolib.core.base.models import ModelSaveSettings
            >>> from hydrolib.core.dflowfm.t3d.serializer import T3DSerializer, T3DSerializerConfig
            >>> data = {
            ... "comments": [],
            ... "layer_type": "SIGMA",
            ... "layers": [0.0, 0.2, 0.6, 0.8, 1.0],
            ... "records": [
            ...         {
            ...             "time": "0 seconds since 2006-01-01 00:00:00 +00:00",
            ...             "data": [1.0, 1.0, 1.0, 1.0, 1.0],
            ...         }
            ...     ],
            ... }
            >>> output_path = Path("tests/data/output/test_serialize.t3d")
            >>> config = T3DSerializerConfig(float_format=".1f")
            >>> T3DSerializer.serialize(output_path, data, config, ModelSaveSettings())
            >>> # Check the contents of the file
            >>> with output_path.open("r") as f:
            ...     print(f.read())
            LAYER_TYPE=SIGMA
            LAYERS=0.0 0.2 0.6 0.8 1.0
            TIME = 0 seconds since 2006-01-01 00:00:00 +00:00
            1.0 1.0 1.0 1.0 1.0
            <BLANKLINE>

            ```
        """
        format_float = lambda v: f"{v:{config.float_format}}"

        with path.open("w", encoding="utf-8") as f:

            # 1) Write the header comments (each line is prefixed by `#`)
            for comment in data.get("comments", []):
                # If the line is blank, just write a blank line; otherwise prefix with '#'
                if comment.strip():
                    f.write(f"#{comment.strip()}\n")
                else:
                    f.write("\n")

            # 2) Write layer type if present, e.g. LAYER_TYPE=SIGMA
            layer_type = data.get("layer_type", None)
            if layer_type:
                f.write(f"LAYER_TYPE={layer_type}\n")

            # 3) Write layers if present, e.g. LAYERS=0.0 0.25 0.75 1.0
            layers = data.get("layers", [])
            if layers:
                layer_str = " ".join(format_float(x) for x in layers)
                f.write(f"LAYERS={layer_str}\n")

            # 4) If VECTORMAX is defined, write it (this is optional in the T3D format).
            vectormax = data.get("vectormax", None)
            if vectormax is not None:
                f.write(f"VECTORMAX={vectormax}\n")

            # 5) Write each record. For each record:
            #      TIME = <time string>
            #      <float1> <float2> ...
            for record in data.get("records", []):
                time_val = record.get("time", "")
                f.write(f"TIME = {time_val}\n")
                data_values = record.get("data", [])
                # Format each float, then join with a space
                line = " ".join(format_float(v) for v in data_values)
                f.write(line + "\n")



from pathlib import Path
from typing import Annotated, Any, Dict, List, Literal, Optional, Set, Union

from pydantic import (
    BeforeValidator,
    ConfigDict,
    Field,
    ValidationInfo,
    field_validator,
    model_validator,
)
from strenum import StrEnum

from hydrolib.core.base.models import (
    DiskOnlyFileModel,
    set_default_disk_only_file_model,
)
from hydrolib.core.base.utils import resolve_file_model, str_is_empty_or_none
from hydrolib.core.dflowfm.bc.models import (
    ForcingBase,
    ForcingData,
    ForcingModel,
    RealTime,
)
from hydrolib.core.dflowfm.common.models import Operand
from hydrolib.core.dflowfm.ini.models import INIBasedModel, INIGeneral, INIModel
from hydrolib.core.dflowfm.ini.serializer import INISerializerConfig
from hydrolib.core.dflowfm.ini.util import (
    LocationValidationConfiguration,
    UnknownKeywordErrorManager,
    enum_value_parser,
    make_list,
    split_string_on_delimiter,
    validate_location_specification,
)
from hydrolib.core.dflowfm.polyfile.models import PolyFile
from hydrolib.core.dflowfm.tim.models import TimModel

SOURCE_SINKS_QUANTITIES_VALID_PREFIXES = (
    "initialtracer",
    "tracerbnd",
    "sedfracbnd",
    "initialsedfrac",
)


FILETYPE_FILEMODEL_MAPPING = {
    "bcascii": ForcingModel,
    "uniform": TimModel,
    "unimagdir": TimModel,
    "arcinfo": DiskOnlyFileModel,
    "spiderweb": DiskOnlyFileModel,
    "curvigrid": DiskOnlyFileModel,
    "netcdf": DiskOnlyFileModel,
    "polygon": PolyFile,
}


class Boundary(INIBasedModel):
    """
    A `[Boundary]` block for use inside an external forcings file,
    i.e., a [ExtModel][hydrolib.core.dflowfm.ext.models.ExtModel].

    All lowercased attributes match with the boundary input as described in
    [UM Sec.C.5.2.1](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.5.2.1).
    """

    _header: Literal["Boundary"] = "Boundary"
    quantity: str = Field(alias="quantity")
    nodeid: Optional[str] = Field(None, alias="nodeId")
    locationfile: Annotated[
        DiskOnlyFileModel, BeforeValidator(set_default_disk_only_file_model)
    ] = Field(default_factory=lambda: DiskOnlyFileModel(None), alias="locationFile")
    forcingfile: List[ForcingModel] = Field(alias="forcingFile")
    bndwidth1d: Optional[float] = Field(None, alias="bndWidth1D")
    bndbldepth: Optional[float] = Field(None, alias="bndBlDepth")
    returntime: Optional[float] = Field(None, alias="returnTime")

    def is_intermediate_link(self) -> bool:
        return True

    @field_validator("forcingfile", mode="before")
    @classmethod
    def validate_forcingfile(cls, data: Any) -> Any:
        if not isinstance(data, list):
            data = [data]
        for i, forcing in enumerate(data):
            if isinstance(forcing, (str, Path)):
                data[i] = ForcingModel(filepath=forcing)
            elif not isinstance(forcing, ForcingModel):
                raise TypeError(
                    "Forcing file must be a ForcingModel or a path to a forcing file."
                )
        return data

    @classmethod
    def _is_valid_locationfile_data(
        cls, elem: Union[None, str, Path, DiskOnlyFileModel]
    ) -> bool:
        return isinstance(elem, Path) or (
            isinstance(elem, DiskOnlyFileModel) and elem.filepath is not None
        )

    @classmethod
    def _exclude_from_validation(cls, input_data: Optional[dict] = None) -> Set:
        unknown_keywords = ["return_time"]
        return set(unknown_keywords)

    @model_validator(mode="before")
    @classmethod
    def rename_return_time_field(cls, values: Dict) -> Dict:
        """Renames the deprecated return_time field to returnTime.

        Args:
            values (Dict): Dictionary with raw, unvalidated input values.

        Returns:
            Dict: Validated dictionary of values for Boundary.
        """
        if "return_time" in values:
            values["returnTime"] = values.pop("return_time")
        return values

    @model_validator(mode="before")
    @classmethod
    def check_nodeid_or_locationfile_present(cls, values: Dict) -> Dict:
        """Verifies that either nodeid or locationfile properties have been set.

        Args:
            values (Dict): Dictionary with values already validated.

        Raises:
            ValueError: When none of the values are present.

        Returns:
            Dict: Validated dictionary of values for Boundary.
        """
        node_id = values.get("nodeid")
        location_file = values.get("locationfile")
        if str_is_empty_or_none(node_id) and not cls._is_valid_locationfile_data(
            location_file
        ):
            raise ValueError(
                "Either nodeId or locationFile fields should be specified."
            )
        return values

    def _get_identifier(self, data: dict) -> Optional[str]:
        """
        Retrieves the identifier for a boundary, which is the nodeid

        Args:
            data (dict): Dictionary of values for this boundary.

        Returns:
            str: The nodeid value or None if not found.
        """
        return data.get("nodeid")

    @property
    def forcing(self) -> Union[ForcingBase, None]:
        """Retrieves the corresponding forcing data for this boundary.

        Returns:
            ForcingBase: The corresponding forcing data. None when this boundary does not have a forcing file or when the data cannot be found.
        """
        result = None
        if self.forcingfile is not None:
            for forcing in self.forcingfile[0].forcing:

                if self.nodeid == forcing.name and any(
                    quantity.quantity.startswith(self.quantity)
                    for quantity in forcing.quantityunitpair
                ):
                    result = forcing
                    break

        return result

    @model_validator(mode="before")
    @classmethod
    def validate_locationfile(cls, data: Any) -> Any:
        file_location = data.get("locationfile") or data.get("locationFile")
        data.pop("locationFile", None)  # Remove alias if present

        # Convert string to DiskOnlyFileModel if needed
        if isinstance(file_location, (str, Path)):
            data["locationfile"] = DiskOnlyFileModel(file_location)
        return data


class Lateral(INIBasedModel):
    """
    A `[Lateral]` block for use inside an external forcings file,
    i.e., a [ExtModel][hydrolib.core.dflowfm.ext.models.ExtModel].

    All lowercased attributes match with the lateral input as described in
    [UM Sec.C.5.2.2](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.5.2.2).
    """

    _header: Literal["Lateral"] = "Lateral"
    id: str = Field(alias="id")
    name: str = Field("", alias="name")
    locationtype: Optional[str] = Field(None, alias="locationType")
    nodeid: Optional[str] = Field(None, alias="nodeId")
    branchid: Optional[str] = Field(None, alias="branchId")
    chainage: Optional[float] = Field(None, alias="chainage")
    numcoordinates: Optional[int] = Field(None, alias="numCoordinates")
    xcoordinates: Optional[List[float]] = Field(None, alias="xCoordinates")
    ycoordinates: Optional[List[float]] = Field(None, alias="yCoordinates")
    discharge: ForcingData = Field(alias="discharge")

    def is_intermediate_link(self) -> bool:
        return True

    @field_validator("xcoordinates", "ycoordinates", mode="before")
    @classmethod
    def split_coordinates(cls, v, info: ValidationInfo) -> List[float]:
        return split_string_on_delimiter(cls, v, info)

    @field_validator("discharge", mode="before")
    @classmethod
    def validate_discharge(cls, v):
        result = v
        if isinstance(v, str):
            # Try float
            try:
                result = float(v)
            except ValueError:
                # Try RealTime enum (case-insensitive)
                try:
                    result = RealTime(v.lower())
                except ValueError:
                    result = resolve_file_model(v, ForcingModel)
        elif isinstance(v, Path):
            result = resolve_file_model(v, ForcingModel)
        elif isinstance(v, dict):
            # Try to instantiate ForcingModel from dict
            result = ForcingModel(**v)
        return result

    @model_validator(mode="before")
    def validate_that_location_specification_is_correct(cls, values: Dict) -> Dict:
        """Validates that the correct location specification is given."""
        return validate_location_specification(
            values, config=LocationValidationConfiguration(minimum_num_coordinates=1)
        )

    def _get_identifier(self, data: dict) -> Optional[str]:
        return data.get("id") or data.get("name")

    @field_validator("locationtype", mode="before")
    @classmethod
    def validate_location_type(cls, v: str) -> str:
        """
        Method to validate whether the specified location type is correct.

        Args:
            v (str): Given value for the locationtype field.

        Raises:
            ValueError: When the value given for locationtype is unknown.

        Returns:
            str: Validated locationtype string.
        """
        possible_values = ["1d", "2d", "all"]
        if v.lower() not in possible_values:
            raise ValueError(
                "Value given ({}) not accepted, should be one of: {}".format(
                    v, ", ".join(possible_values)
                )
            )
        return v


class SourceSink(INIBasedModel):
    """
    A `[SourceSink]` block for use inside an external forcings file,
    i.e., a [ExtModel][hydrolib.core.dflowfm.ext.models.SourceSink].

    All lowercased attributes match with the source-sink input as described in
    [UM Sec.C.5.2.4](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.5.2.4).
    """

    _header: Literal["SourceSink"] = "SourceSink"
    id: str = Field(alias="id")
    name: str = Field("", alias="name")
    locationfile: Optional[DiskOnlyFileModel] = Field(
        default_factory=lambda: DiskOnlyFileModel(None), alias="locationFile"
    )

    numcoordinates: Optional[int] = Field(None, alias="numCoordinates")
    xcoordinates: Optional[List[float]] = Field(None, alias="xCoordinates")
    ycoordinates: Optional[List[float]] = Field(None, alias="yCoordinates")

    zsource: Optional[Union[float, List[float]]] = Field(None, alias="zSource")
    zsink: Optional[Union[float, List[float]]] = Field(None, alias="zSink")
    area: Optional[float] = Field(None, alias="Area")

    discharge: ForcingData = Field(alias="discharge")
    salinitydelta: Optional[ForcingData] = Field(None, alias="salinityDelta")
    temperaturedelta: Optional[ForcingData] = Field(None, alias="temperatureDelta")

    def is_intermediate_link(self) -> bool:
        return True

    @classmethod
    def _exclude_from_validation(cls, input_data: Optional[dict] = None) -> Set:
        fields = cls.model_fields
        unknown_keywords = [
            key
            for key in input_data.keys()
            if key not in fields
            and key.startswith(SOURCE_SINKS_QUANTITIES_VALID_PREFIXES)
        ]
        return set(unknown_keywords)

    model_config = ConfigDict(extra="allow")

    def __init__(self, **data):
        super().__init__(**data)
        # Add dynamic attributes for fields starting with 'tracer'
        for key, value in data.items():
            if isinstance(key, str) and key.startswith(
                SOURCE_SINKS_QUANTITIES_VALID_PREFIXES
            ):
                setattr(self, key, value)

    @model_validator(mode="before")
    def validate_location_specification(cls, values):
        """
        Ensures that either `locationfile` or a valid set of coordinates is provided.

         This validation enforces that at least one of the following conditions is met:
         1. `locationfile` is provided.
         2. The combination of `numcoordinates`, `xcoordinates`, and `ycoordinates` is valid:
             - `xcoordinates` and `ycoordinates` must be lists of equal length.
             - The length of `xcoordinates` and `ycoordinates` must match `numcoordinates`.

         Raises:
             ValueError: If neither `locationfile` nor a valid coordinate set is provided.

         Returns:
             Dict: The validated input values.
        """
        locationfile = values.get("locationfile", values.get("locationFile"))

        numcoordinates = values.get("numcoordinates", values.get("numCoordinates"))
        xcoordinates = values.get("xcoordinates", values.get("xCoordinates"))
        ycoordinates = values.get("ycoordinates", values.get("yCoordinates"))

        has_locationfile = locationfile is not None
        has_coordinates = (
            numcoordinates is not None
            and xcoordinates is not None
            and ycoordinates is not None
            and len(xcoordinates) == len(ycoordinates) == numcoordinates
        )

        if not (has_locationfile or has_coordinates):
            raise ValueError(
                "Either `locationFile` or the combination of `numCoordinates`, `xCoordinates`, and `yCoordinates` "
                f"must be provided. for the SourceSink block `{values.get('id')}`."
            )

        return values

    @model_validator(mode="before")
    @classmethod
    def validate_locationfile(cls, data: Any) -> Any:
        file_location = data.get("locationfile") or data.get("locationFile")
        data.pop("locationFile", None)  # Remove alias if present

        # Convert string to DiskOnlyFileModel if needed
        if isinstance(file_location, (str, Path)):
            data["locationfile"] = DiskOnlyFileModel(file_location)
        else:
            data["locationfile"] = file_location
        return data


class MeteoForcingFileType(StrEnum):
    """
    Enum class containing the valid values for the forcingFileType
    attribute in Meteo class.
    """

    bcascii = "bcAscii"
    """str: Space-uniform time series in <*.bc> file."""

    uniform = "uniform"
    """str: Space-uniform time series in <*.tim> file."""

    unimagdir = "uniMagDir"
    """str: Space-uniform wind magnitude+direction in <*.tim> file."""

    arcinfo = "arcInfo"
    """str: Space- and time-varying wind and pressure on an equidistant grid in <*.amu/v/p> files."""

    spiderweb = "spiderweb"
    """str: Space- and time-varying cyclone wind and pressure in <*.spw> files."""

    curvigrid = "curviGrid"
    """str: Space- and time-varying wind and pressure on a curvilinear grid in <*.grd+*.amu/v/p> files."""

    netcdf = "netcdf"
    """str: NetCDF, either with gridded data, or multiple station time series."""

    polygon = "polygon"
    """str: Polygon-based time series in <*.pol> file."""

    allowedvaluestext = "Possible values: bcAscii, uniform, uniMagDir, arcInfo, spiderweb, curviGrid, netcdf, polygon."


class MeteoInterpolationMethod(StrEnum):
    """
    Enum class containing the valid values for the interpolationMethod
    attribute in Meteo class.
    """

    nearestnb = "nearestNb"
    """str: Nearest-neighbour interpolation, only with station-data in forcingFileType=netcdf"""
    linearSpaceTime = "linearSpaceTime"
    """str: Linear interpolation in space and time."""
    constant = "constant"
    allowedvaluestext = "Possible values: nearestNb, linearSpaceTime, constant."


class Meteo(INIBasedModel):
    """
    A `[Meteo]` block for use inside an external forcings file,
    i.e., a [ExtModel][hydrolib.core.dflowfm.ext.models.ExtModel].

    All lowercased attributes match with the meteo input as described in
    [UM Sec.C.5.2.3](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.5.2.3).
    """

    class Comments(INIBasedModel.Comments):
        quantity: Optional[str] = Field(
            "Name of the quantity. See UM Section C.5.3", alias="quantity"
        )
        forcingfile: Optional[str] = Field(
            "Name of file containing the forcing for this meteo quantity.",
            alias="forcingFile",
        )
        forcingfiletype: Optional[str] = Field(
            "Type of forcingFile.", alias="forcingFileType"
        )
        forcingVariableName: Optional[str] = Field(
            "Variable name used in forcingfile associated with this forcing. See UM Section C.5.3",
            alias="forcingVariableName",
        )
        targetmaskfile: Optional[str] = Field(
            "Name of <*.pol> file to be used as mask. Grid parts inside any polygon will receive the meteo forcing.",
            alias="targetMaskFile",
        )
        targetmaskinvert: Optional[str] = Field(
            "Flag indicating whether the target mask should be inverted, i.e., outside of all polygons: no or yes.",
            alias="targetMaskInvert",
        )
        interpolationmethod: Optional[str] = Field(
            "Type of (spatial) interpolation.", alias="interpolationMethod"
        )
        operand: Optional[str] = Field(
            "How this data is combined with previous data for the same quantity (if any).",
            alias="operand",
        )
        extrapolationAllowed: Optional[str] = Field(
            "Optionally allow nearest neighbour extrapolation in space (0: no, 1: yes). Default off.",
            alias="extrapolationAllowed",
        )
        extrapolationSearchRadius: Optional[str] = Field(
            "Maximum search radius for nearest neighbor extrapolation in space.",
            alias="extrapolationSearchRadius",
        )

    comments: Comments = Comments()

    @classmethod
    def _get_unknown_keyword_error_manager(cls) -> Optional[UnknownKeywordErrorManager]:
        """
        The Meteo does not currently support raising an error on unknown keywords.
        """
        return None

    _header: Literal["Meteo"] = "Meteo"
    quantity: str = Field(alias="quantity")
    forcingfile: Union[TimModel, ForcingModel, DiskOnlyFileModel, PolyFile] = Field(
        alias="forcingFile"
    )
    forcingVariableName: Optional[str] = Field(None, alias="forcingVariableName")
    forcingfiletype: MeteoForcingFileType = Field(alias="forcingFileType")
    targetmaskfile: Optional[PolyFile] = Field(None, alias="targetMaskFile")
    targetmaskinvert: Optional[bool] = Field(None, alias="targetMaskInvert")
    interpolationmethod: Optional[MeteoInterpolationMethod] = Field(
        None, alias="interpolationMethod"
    )
    operand: Optional[Operand] = Field(Operand.override.value, alias="operand")
    extrapolationAllowed: Optional[bool] = Field(None, alias="extrapolationAllowed")
    extrapolationSearchRadius: Optional[float] = Field(
        None, alias="extrapolationSearchRadius"
    )
    averagingType: Optional[int] = Field(None, alias="averagingType")
    averagingNumMin: Optional[float] = Field(None, alias="averagingNumMin")
    averagingPercentile: Optional[float] = Field(None, alias="averagingPercentile")

    @model_validator(mode="before")
    @classmethod
    def choose_file_model(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Root-level validator to the right class for the filename parameter based on the filetype.

        The validator chooses the right class for the filename parameter based on the FileType_FileModel_mapping
        dictionary.

        FILETYPE_FILEMODEL_MAPPING = {
            "bcascii": ForcingModel,
            "uniform": TimModel,
            "unimagdir": TimModel,
            "arcinfo": DiskOnlyFileModel,
            "spiderweb": DiskOnlyFileModel,
            "curvigrid": DiskOnlyFileModel,
            "netcdf": DiskOnlyFileModel,
            "polygon": PolyFile,
        }
        """
        # if the filetype and the filename are present in the values
        if any(par in values for par in ["forcingfiletype", "forcingFileType"]) and any(
            par in values for par in ["forcingfile", "forcingFile"]
        ):
            file_type_var_name = (
                "forcingfiletype" if "forcingfiletype" in values else "forcingFileType"
            )
            filename_var_name = (
                "forcingfile" if "forcingfile" in values else "forcingFile"
            )
            file_type = values.get(file_type_var_name)
            file_type = str(file_type).lower() if file_type is not None else None
            raw_path = values.get(filename_var_name)
            if isinstance(raw_path, (Path, str)):
                model = FILETYPE_FILEMODEL_MAPPING.get(file_type)
                values[filename_var_name] = resolve_file_model(raw_path, model)

        return values

    def is_intermediate_link(self) -> bool:
        return True

    @field_validator("forcingfiletype", mode="before")
    @classmethod
    def forcingfiletype_validator(cls, v):
        return enum_value_parser(v, MeteoForcingFileType)

    @field_validator("interpolationmethod", mode="before")
    @classmethod
    def interpolationmethod_validator(cls, v):
        return enum_value_parser(v, MeteoInterpolationMethod)


class ExtGeneral(INIGeneral):
    """The external forcing file's `[General]` section with file meta-data."""

    _header: Literal["General"] = "General"
    fileversion: str = Field("2.01", alias="fileVersion")
    filetype: Literal["extForce"] = Field("extForce", alias="fileType")


class ExtModel(INIModel):
    """
    The overall external forcings model that contains the contents of one external forcings file (new format).

    This model is typically referenced under a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.external_forcing.extforcefilenew`.

    Attributes:
        general (ExtGeneral): `[General]` block with file metadata.
        boundary (List[Boundary]): List of `[Boundary]` blocks for all boundary conditions.
        lateral (List[Lateral]): List of `[Lateral]` blocks for all lateral discharges.
        sourcesink (List[SourceSink]): List of `[SourceSink]` blocks for all source/sink terms.
        meteo (List[Meteo]): List of `[Meteo]` blocks for all meteorological forcings.
    """

    general: ExtGeneral = ExtGeneral()
    boundary: Annotated[List[Boundary], BeforeValidator(make_list)] = Field(
        default_factory=list
    )
    lateral: Annotated[List[Lateral], BeforeValidator(make_list)] = Field(
        default_factory=list
    )
    sourcesink: Annotated[List[SourceSink], BeforeValidator(make_list)] = Field(
        default_factory=list
    )
    meteo: Annotated[List[Meteo], BeforeValidator(make_list)] = Field(
        default_factory=list
    )
    serializer_config: INISerializerConfig = INISerializerConfig(
        section_indent=0, property_indent=0
    )

    @classmethod
    def _ext(cls) -> str:
        return ".ext"

    @classmethod
    def _filename(cls) -> str:
        return "bnd"


class SourceSinkError(Exception):
    """SourceSinkError."""

    def __init__(self, error_message: str):
        """SourceSinkError constructor."""
        print(error_message)


class InitialFieldError(Exception):
    """InitialFieldError."""

    def __init__(self, error_message: str):
        """InitialFieldError constructor."""
        print(error_message)


class MeteoError(Exception):
    """MeteoError."""

    def __init__(self, error_message: str):
        """MeteoError constructor."""
        print(error_message)


class BoundaryError(Exception):
    """BoundaryError."""

    def __init__(self, error_message: str):
        """BoundaryError constructor."""
        print(error_message)


class ParameterFieldError(Exception):
    """ParameterFieldError."""

    def __init__(self, error_message: str):
        """ParameterFieldError constructor."""
        print(error_message)



"""Structure models for D-Flow FM.

structure namespace for storing the contents of an [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]'s structure file.
"""

import logging
from enum import Enum
from operator import gt, ne
from pathlib import Path
from typing import Annotated, Dict, List, Literal, Optional, Set, Union

from pydantic import (
    BeforeValidator,
    Field,
    ValidationInfo,
    field_validator,
    model_validator,
)
from strenum import StrEnum

from hydrolib.core.base.models import DiskOnlyFileModel
from hydrolib.core.base.utils import str_is_empty_or_none
from hydrolib.core.dflowfm.bc.models import ForcingModel
from hydrolib.core.dflowfm.friction.models import FrictionType
from hydrolib.core.dflowfm.ini.models import INIBasedModel, INIGeneral, INIModel
from hydrolib.core.dflowfm.ini.util import (
    UnknownKeywordErrorManager,
    enum_value_parser,
    make_list,
    split_string_on_delimiter,
    validate_conditionally,
    validate_correct_length,
    validate_forbidden_fields,
    validate_required_fields,
)
from hydrolib.core.dflowfm.tim.models import TimModel

logger = logging.getLogger(__name__)


def load_model(value):
    """Load a model from a file path or return the value as is."""
    if isinstance(value, (str, Path)):
        path = Path(value)
        if path.suffix == ".tim":
            return TimModel(filepath=path)
        elif path.suffix == ".bc":
            return ForcingModel(filepath=path)
    return value  # already a float, TimModel, or ForcingModel


ForcingDataUnion = Annotated[
    Union[float, TimModel, ForcingModel], BeforeValidator(load_model)
]


class Structure(INIBasedModel):
    """Structure model."""

    class Comments(INIBasedModel.Comments):
        id: Optional[str] = "Unique structure id (max. 256 characters)."
        name: Optional[str] = "Given name in the user interface."
        polylinefile: Optional[str] = Field(
            "*.pli; Polyline geometry definition for 2D structure.",
            alias="polylinefile",
        )
        branchid: Optional[str] = Field(
            "Branch on which the structure is located.", alias="branchId"
        )
        chainage: Optional[str] = "Chainage on the branch (m)."

        numcoordinates: Optional[str] = Field(
            "Number of values in xCoordinates and yCoordinates", alias="numCoordinates"
        )
        xcoordinates: Optional[str] = Field(
            "x-coordinates of the location of the structure. (number of values = numCoordinates)",
            alias="xCoordinates",
        )
        ycoordinates: Optional[str] = Field(
            "y-coordinates of the location of the structure. (number of values = numCoordinates)",
            alias="yCoordinates",
        )

    comments: Comments = Comments()

    _header: Literal["Structure"] = "Structure"

    id: str = Field("id", max_length=256, alias="id")
    name: str = Field("id", alias="name")
    type: str = Field(alias="type")

    polylinefile: Optional[DiskOnlyFileModel] = Field(None, alias="polylinefile")

    branchid: Optional[str] = Field(None, alias="branchId")
    chainage: Optional[float] = Field(None, alias="chainage")

    numcoordinates: Optional[int] = Field(None, alias="numCoordinates")
    xcoordinates: Optional[List[float]] = Field(None, alias="xCoordinates")
    ycoordinates: Optional[List[float]] = Field(None, alias="yCoordinates")

    _loc_coord_fields = {"numcoordinates", "xcoordinates", "ycoordinates"}
    _loc_branch_fields = {"branchid", "chainage"}
    _loc_all_fields = _loc_coord_fields | _loc_branch_fields

    @classmethod
    def _get_unknown_keyword_error_manager(cls) -> Optional[UnknownKeywordErrorManager]:
        """Get the UnknownKeywordErrorManager for this model.

        The Structure does not currently support raising an error on unknown keywords.
        """
        return None

    @field_validator("xcoordinates", "ycoordinates", mode="before")
    @classmethod
    def _split_to_list(cls, v, info: ValidationInfo):
        return split_string_on_delimiter(cls, v, info)

    @field_validator("type", mode="after")
    @classmethod
    def lowercase_type(cls, v: str) -> str:
        if isinstance(v, str):
            return v.lower()
        return v

    @model_validator(mode="after")
    def check_location(self):
        """Check location of the structure.

        Validates the location of the structure based on the given parameters for the following cases:
            - if a branchid is given, then it is expected also the chainage, otherwise numcoordinates xcoordinates
            and ycoordinates shall be expected.

        Raises:
            ValueError: When branchid or chainage values are not valid (empty strings).
            ValueError: When the number of xcoordinates and ycoordinates do not match numcoordinates.

        Returns:
            dict: Dictionary of values validated for the new structure.
        """
        values_dict = self.model_dump()

        filtered_values = {k: v for k, v in values_dict.items() if v is not None}
        structype = filtered_values.get("type", "").lower()

        if structype == "compound" or issubclass(self.__class__, Compound):
            # Compound structure does not require a location specification.
            return self

        # Backwards compatibility for old-style polylinefile input field (instead of num/x/yCoordinates):
        polyline_compatible_structures = dict(
            pump="Pump",
            dambreak="Dambreak",
            gate="Gate",
            weir="Weir",
            generalstructure="GeneralStructure",
        )
        polylinefile_in_model = (
            structype in polyline_compatible_structures.keys()
            and filtered_values.get("polylinefile") is not None
        )

        # No branchId+chainage for some structures:
        only_coordinates_structures = dict(
            longculvert="LongCulvert", dambreak="Dambreak"
        )
        coordinates_in_model = Structure.validate_coordinates_in_model(filtered_values)

        # Error: do not allow both x/y and polyline file:
        if polylinefile_in_model and coordinates_in_model:
            raise ValueError(
                "Specify location either by `num/x/yCoordinates` or `polylinefile`, but not both."
            )

        # Error: require x/y or polyline file:
        if (
            structype in polyline_compatible_structures.keys()
            and structype in only_coordinates_structures.keys()
        ):
            if not (coordinates_in_model or polylinefile_in_model):
                raise ValueError(
                    f"Specify location either by setting `num/x/yCoordinates` or `polylinefile` fields for a {polyline_compatible_structures[structype]} structure."
                )

        # Error: Some structures require coordinates_in_model, but not branchId and chainage.
        if (
            not polylinefile_in_model
            and structype in only_coordinates_structures.keys()
        ):
            if not coordinates_in_model:
                raise ValueError(
                    f"Specify location by setting `num/x/yCoordinates` for a {only_coordinates_structures[structype]} structure."
                )

        # Error: final check: at least one of x/y, branchId+chainage or polyline file must be given
        branch_and_chainage_in_model = Structure.validate_branch_and_chainage_in_model(
            filtered_values
        )

        if not (
            branch_and_chainage_in_model
            or coordinates_in_model
            or polylinefile_in_model
        ):
            raise ValueError(
                "Specify location either by setting `branchId` and `chainage` or `num/x/yCoordinates` or `polylinefile` fields."
            )

        return self

    @field_validator("polylinefile", mode="before")
    @classmethod
    def resolve_polylinefile(cls, value) -> dict:
        if isinstance(value, (str, Path)):
            return DiskOnlyFileModel(filepath=Path(value))
        return value

    @staticmethod
    def validate_branch_and_chainage_in_model(values: dict) -> bool:
        """Validate branch and chainage in model.

        Static method to validate whether the given branchid and chainage values
        match the expectation of a new structure.

        Args:
            values (dict): Dictionary of values to be used to generate a structure.

        Raises:
            ValueError: When the value for branchid or chainage are not valid.

        Returns:
            bool: Result of valid branchid / chainage in dictionary.
        """
        branchid = values.get("branchid", None)
        if branchid is None:
            return False

        chainage = values.get("chainage", None)
        if str_is_empty_or_none(branchid) or chainage is None:
            raise ValueError(
                "A valid value for branchId and chainage is required when branchId key is specified."
            )
        return True

    @staticmethod
    def validate_coordinates_in_model(values: dict) -> bool:
        """Validate coordinates in model.

        Static method to validate whether the given values match the expectations
        of a structure to define its coordinates.

        Args:
            values (dict): Dictionary of values to be used to generate a structure.

        Raises:
            ValueError: When the given coordinates is less than 2.
            ValueError: When the given coordinates do not match in expected size.

        Returns:
            bool: Result of valid coordinates in dictionary.
        """
        searched_keys = ["numcoordinates", "xcoordinates", "ycoordinates"]
        if any(values.get(k, None) is None for k in searched_keys):
            return False

        n_coords = values["numcoordinates"]
        if n_coords < 2:
            raise ValueError(
                f"Expected at least 2 coordinates, but only {n_coords} declared."
            )

        def get_coord_len(coord: str) -> int:
            if values[coord] is None:
                return 0
            return len(values[coord])

        len_x_coords = get_coord_len("xcoordinates")
        len_y_coords = get_coord_len("ycoordinates")
        if n_coords == len_x_coords == len_y_coords:
            return True
        raise ValueError(
            f"Expected {n_coords} coordinates, given {len_x_coords} for xCoordinates and {len_y_coords} for yCoordinates."
        )

    def _exclude_fields(self) -> Set:
        # exclude the non-applicable, or unset props like coordinates or branches
        if self.type == "compound":
            exclude_set = self._loc_all_fields
        elif self.branchid is not None:
            exclude_set = self._loc_coord_fields
        else:
            exclude_set = self._loc_branch_fields
        exclude_set = super()._exclude_fields().union(exclude_set)
        return exclude_set

    def _get_identifier(self, data: dict) -> Optional[str]:
        return data.get("id") or data.get("name")


class FlowDirection(StrEnum):
    """Flow direction of a structure.

    Enum class containing the valid values for the allowedFlowDirection
    attribute in several subclasses of Structure.
    """

    none = "none"
    positive = "positive"
    negative = "negative"
    both = "both"
    allowedvaluestext = "Possible values: both, positive, negative, none."


class Orientation(StrEnum):
    """Orientation of a structure.

    Enum class containing the valid values for the orientation
    attribute in several subclasses of Structure.
    """

    positive = "positive"
    negative = "negative"
    allowedvaluestext = "Possible values: positive, negative."


class Weir(Structure):
    """Weir structure.

    Hydraulic structure with `type=weir`, to be included in a structure file.
    Typically inside the structure list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.structurefile[0].structure[..]`

    All lowercased attributes match with the weir input as described in
    [UM Sec.C.12.1](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.12.1).
    """

    class Comments(Structure.Comments):
        type: Optional[str] = Field("Structure type; must read weir", alias="type")
        allowedflowdir: Optional[str] = Field(
            FlowDirection.allowedvaluestext, alias="allowedFlowdir"
        )

        crestlevel: Optional[str] = Field(
            "Crest level of weir (m AD).", alias="crestLevel"
        )
        crestwidth: Optional[str] = Field("Width of the weir (m).", alias="crestWidth")
        corrcoeff: Optional[str] = Field(
            "Correction coefficient (-).", alias="corrCoeff"
        )
        usevelocityheight: Optional[str] = Field(
            "Flag indicating whether the velocity height is to be calculated or not.",
            alias="useVelocityHeight",
        )

    comments: Comments = Comments()

    type: Literal["weir"] = Field("weir", alias="type")
    allowedflowdir: Optional[FlowDirection] = Field(
        FlowDirection.both.value, alias="allowedFlowDir"
    )

    crestlevel: ForcingDataUnion = Field(alias="crestLevel")
    crestwidth: Optional[float] = Field(None, alias="crestWidth")
    corrcoeff: float = Field(1.0, alias="corrCoeff")
    usevelocityheight: bool = Field(True, alias="useVelocityHeight")

    @field_validator("allowedflowdir", mode="before")
    @classmethod
    def _flowdirection_validator(cls, v) -> FlowDirection:
        return enum_value_parser(v, FlowDirection)


class UniversalWeir(Structure):
    """Universal Weir structure.

    Hydraulic structure with `type=universalWeir`, to be included in a structure file.
    Typically inside the structure list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.structurefile[0].structure[..]`

    All lowercased attributes match with the universal weir input as described in
    [UM Sec.C.12.2](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.12.2).
    """

    class Comments(Structure.Comments):
        type: Optional[str] = Field(
            "Structure type; must read universalWeir", alias="type"
        )
        allowedflowdir: Optional[str] = Field(
            FlowDirection.allowedvaluestext, alias="allowedFlowdir"
        )

        numlevels: Optional[str] = Field("Number of yz-Values.", alias="numLevels")
        yvalues: Optional[str] = Field(
            "y-values of the cross section (m). (number of values = numLevels)",
            alias="yValues",
        )
        zvalues: Optional[str] = Field(
            "z-values of the cross section (m). (number of values = numLevels)",
            alias="zValues",
        )
        crestlevel: Optional[str] = Field(
            "Crest level of weir (m AD).", alias="crestLevel"
        )
        dischargecoeff: Optional[str] = Field(
            "Discharge coefficient c_e (-).", alias="dischargeCoeff"
        )

    comments: Comments = Comments()

    type: Literal["universalWeir"] = Field("universalWeir", alias="type")
    allowedflowdir: FlowDirection = Field(alias="allowedFlowDir")

    numlevels: int = Field(alias="numLevels")
    yvalues: List[float] = Field(alias="yValues")
    zvalues: List[float] = Field(alias="zValues")
    crestlevel: float = Field(alias="crestLevel")
    dischargecoeff: float = Field(alias="dischargeCoeff")

    @field_validator("yvalues", "zvalues", mode="before")
    @classmethod
    def _split_to_list(cls, v, info: ValidationInfo):
        return split_string_on_delimiter(cls, v, info)

    @field_validator("allowedflowdir", mode="before")
    @classmethod
    def _flowdirection_validator(cls, v) -> FlowDirection:
        return enum_value_parser(v, FlowDirection)


class CulvertSubType(StrEnum):
    """Enum class to store a [Culvert][hydrolib.core.dflowfm.structure.models.Culvert]'s subType."""

    culvert = "culvert"
    invertedSiphon = "invertedSiphon"


class Culvert(Structure):
    """Culvert structure.

    Hydraulic structure with `type=culvert`, to be included in a structure file.
    Typically inside the structure list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.structurefile[0].structure[..]`

    All lowercased attributes match with the culvert input as described in
    [UM Sec.C.12.3](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.12.3).
    """

    type: Literal["culvert"] = Field("culvert", alias="type")
    allowedflowdir: FlowDirection = Field(alias="allowedFlowDir")

    leftlevel: float = Field(alias="leftLevel")
    rightlevel: float = Field(alias="rightLevel")
    csdefid: str = Field(alias="csDefId")
    length: float = Field(alias="length")
    inletlosscoeff: float = Field(alias="inletLossCoeff")
    outletlosscoeff: float = Field(alias="outletLossCoeff")
    valveonoff: bool = Field(alias="valveOnOff")
    valveopeningheight: Optional[ForcingDataUnion] = Field(
        None, alias="valveOpeningHeight"
    )
    numlosscoeff: Optional[int] = Field(None, alias="numLossCoeff")
    relopening: Optional[List[float]] = Field(None, alias="relOpening")
    losscoeff: Optional[List[float]] = Field(None, alias="lossCoeff")
    bedfrictiontype: Optional[FrictionType] = Field(None, alias="bedFrictionType")
    bedfriction: Optional[float] = Field(None, alias="bedFriction")
    subtype: Optional[CulvertSubType] = Field(
        CulvertSubType.culvert.value, alias="subType"
    )
    bendlosscoeff: Optional[float] = Field(None, alias="bendLossCoeff")

    @field_validator("relopening", "losscoeff", mode="before")
    @classmethod
    def _split_to_list(cls, v, info: ValidationInfo) -> List[float]:
        return split_string_on_delimiter(cls, v, info)

    @field_validator("allowedflowdir", mode="before")
    @classmethod
    def _flowdirection_validator(cls, v) -> FlowDirection:
        return enum_value_parser(v, FlowDirection)

    @field_validator("subtype", mode="before")
    @classmethod
    def _subtype_validator(cls, v) -> CulvertSubType:
        return enum_value_parser(v, CulvertSubType)

    @field_validator("bedfrictiontype", mode="before")
    @classmethod
    def _frictiontype_validator(cls, v) -> FrictionType:
        return enum_value_parser(v, FrictionType)

    @model_validator(mode="after")
    def validate_that_valve_related_fields_are_present_for_culverts_with_valves(self):
        """Validates that valve-related fields are present when there is a valve present."""
        validate_required_fields(
            self.model_dump(),
            "valveopeningheight",
            "numlosscoeff",
            "relopening",
            "losscoeff",
            conditional_field_name="valveonoff",
            conditional_value=True,
        )
        return self

    @model_validator(mode="after")
    def validate_that_bendlosscoeff_field_is_present_for_invertedsyphons(self):
        """Validates that the bendlosscoeff value is present when dealing with inverted syphons."""
        validate_required_fields(
            self.model_dump(),
            "bendlosscoeff",
            conditional_field_name="subtype",
            conditional_value=CulvertSubType.invertedSiphon,
        )
        return self

    @model_validator(mode="after")
    def check_list_lengths(self):
        """Validates that the length of the relopening and losscoeff fields are as expected."""
        validate_correct_length(
            self.model_dump(),
            "relopening",
            "losscoeff",
            length_name="numlosscoeff",
            list_required_with_length=True,
        )
        return self

    @model_validator(mode="after")
    def validate_that_bendlosscoeff_is_not_provided_for_culverts(self):
        """Validates that the bendlosscoeff field is not provided when the subtype is a culvert."""
        validate_forbidden_fields(
            self.model_dump(),
            "bendlosscoeff",
            conditional_field_name="subtype",
            conditional_value=CulvertSubType.culvert,
        )
        return self


class LongCulvert(Structure):
    """Long Culvert structure.

    Hydraulic structure with `type=longCulvert`, to be included in a structure file.
    Typically inside the structure list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.structurefile[0].structure[..]`

    All lowercased attributes match with the long culvert input as described in
    [UM Sec.C.13.4](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.13.4).
    """

    type: Literal["longCulvert"] = Field("longCulvert", alias="type")
    allowedflowdir: Optional[FlowDirection] = Field(None, alias="allowedFlowDir")
    zcoordinates: Optional[List[float]] = Field(None, alias="zCoordinates")

    width: float = Field(alias="width")
    height: float = Field(alias="height")
    frictiontype: FrictionType = Field(alias="frictionType")
    frictionvalue: float = Field(alias="frictionValue")
    valverelativeopening: float = Field(alias="valveRelativeOpening")
    csdefid: Optional[str] = Field(None, alias="csDefId")

    @field_validator("frictiontype", mode="before")
    @classmethod
    def _frictiontype_validator(cls, v) -> FrictionType:
        return enum_value_parser(v, FrictionType)

    @field_validator("allowedflowdir", mode="before")
    @classmethod
    def _flowdirection_validator(cls, v) -> FlowDirection:
        return enum_value_parser(v, FlowDirection)

    @field_validator("zcoordinates", mode="before")
    @classmethod
    def _split_to_list(cls, v, info: ValidationInfo) -> List[float]:
        return split_string_on_delimiter(cls, v, info)

    @field_validator("zcoordinates", mode="after")
    @classmethod
    def _validate_zcoordinates(cls, v, values: ValidationInfo):
        if v is None:
            return v
        num_coordinates = values.data.get("numcoordinates")
        if len(v) != num_coordinates:
            raise ValueError(
                f"Expected {num_coordinates} z-coordinates, but got {len(v)}."
            )

        return v


class Pump(Structure):
    """Pump structure.

    Hydraulic structure with `type=pump`, to be included in a structure file.
    Typically inside the structure list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.structurefile[0].structure[..]`

    All lowercased attributes match with the pump input as described in
    [UM Sec.C.12.6](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.12.6).
    """

    type: Literal["pump"] = Field("pump", alias="type")

    orientation: Optional[Orientation] = Field(None, alias="orientation")
    controlside: Optional[str] = Field(None, alias="controlSide")
    numstages: Optional[int] = Field(None, alias="numStages")
    capacity: ForcingDataUnion = Field(alias="capacity")

    startlevelsuctionside: Optional[List[float]] = Field(
        None, alias="startLevelSuctionSide"
    )
    stoplevelsuctionside: Optional[List[float]] = Field(
        None, alias="stopLevelSuctionSide"
    )
    startleveldeliveryside: Optional[List[float]] = Field(
        None, alias="startLevelDeliverySide"
    )
    stopleveldeliveryside: Optional[List[float]] = Field(
        None, alias="stopLevelDeliverySide"
    )
    numreductionlevels: Optional[int] = Field(None, alias="numReductionLevels")
    head: Optional[List[float]] = Field(None, alias="head")
    reductionfactor: Optional[List[float]] = Field(None, alias="reductionFactor")

    @field_validator(
        "startlevelsuctionside",
        "stoplevelsuctionside",
        "startleveldeliveryside",
        "stopleveldeliveryside",
        "head",
        "reductionfactor",
        mode="before",
    )
    @classmethod
    def _split_to_list(cls, v, info: ValidationInfo) -> List[float]:
        """Split the string on the delimiter and return a list of floats."""
        return split_string_on_delimiter(cls, v, info)

    @field_validator("orientation", mode="before")
    @classmethod
    def _orientation_validator(cls, v) -> Orientation:
        return enum_value_parser(v, Orientation)

    @model_validator(mode="after")
    def validate_that_controlside_is_provided_when_numstages_is_provided(self):
        validate_required_fields(
            self.model_dump(),
            "controlside",
            conditional_field_name="numstages",
            conditional_value=0,
            comparison_func=gt,
        )
        return self

    @classmethod
    def _check_list_lengths_suctionside(cls, values: Dict):
        """Validates that the length of the startlevelsuctionside and stoplevelsuctionside fields are as expected."""
        validate_correct_length(
            values,
            "startlevelsuctionside",
            "stoplevelsuctionside",
            length_name="numstages",
            list_required_with_length=True,
        )

    @model_validator(mode="after")
    def conditionally_check_list_lengths_suctionside(self) -> Dict:
        """Check the length of suction side list.

        Validates the length of the suction side fields, but only if there is a controlside value
        present in the values and the controlside is not equal to the deliverySide.
        """
        validate_conditionally(
            self.model_dump(),
            Pump._check_list_lengths_suctionside,
            "controlside",
            "deliverySide",
            ne,
        )
        return self

    @staticmethod
    def _check_list_lengths_deliveryside(values: Dict) -> Dict:
        """Validates that the length of the startleveldeliveryside and stopleveldeliveryside fields are as expected."""
        validate_correct_length(
            values,
            "startleveldeliveryside",
            "stopleveldeliveryside",
            length_name="numstages",
            list_required_with_length=True,
        )

    @model_validator(mode="after")
    def conditionally_check_list_lengths_deliveryside(self):
        """Check the length of deliveryside list.

        Validates the length of the delivery side fields, but only if there is a controlside value
        present in the values and the controlside is not equal to the suctionSide.
        """
        validate_conditionally(
            self.model_dump(),
            Pump._check_list_lengths_deliveryside,
            "controlside",
            "suctionSide",
            ne,
        )
        return self

    @model_validator(mode="after")
    def check_list_lengths_head_and_reductionfactor(self):
        """Validates that the lengths of the head and reductionfactor fields are as expected."""
        validate_correct_length(
            self.model_dump(),
            "head",
            "reductionfactor",
            length_name="numreductionlevels",
            list_required_with_length=True,
        )
        return self


class Compound(Structure):
    """Compound structure.

    Hydraulic structure with `type=compound`, to be included in a structure file.
    Typically inside the structure list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.structurefile[0].structure[..]`

    All lowercased attributes match with the compound input as described in
    [UM Sec.C.12.11](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.12.11).
    """

    type: Literal["compound"] = Field("compound", alias="type")
    numstructures: int = Field(alias="numStructures")
    structureids: List[str] = Field(alias="structureIds", delimiter=";")

    @field_validator("structureids", mode="before")
    @classmethod
    def _split_to_list(cls, v, info: ValidationInfo) -> List[str]:
        """Split the string on the delimiter and return a list of strings."""
        return split_string_on_delimiter(cls, v, info)


class Orifice(Structure):
    """Orifice structure.

    Hydraulic structure with `type=orifice`, to be included in a structure file.
    Typically inside the structure list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.structurefile[0].structure[..]`

    All lowercased attributes match with the orifice input as described in
    [UM Sec.C.12.7](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.12.7).
    """

    type: Literal["orifice"] = Field("orifice", alias="type")
    allowedflowdir: Optional[FlowDirection] = Field(
        FlowDirection.both.value, alias="allowedFlowDir"
    )

    crestlevel: ForcingDataUnion = Field(alias="crestLevel")
    crestwidth: Optional[float] = Field(None, alias="crestWidth")
    gateloweredgelevel: ForcingDataUnion = Field(alias="gateLowerEdgeLevel")
    corrcoeff: float = Field(1.0, alias="corrCoeff")
    usevelocityheight: bool = Field(True, alias="useVelocityHeight")

    uselimitflowpos: Optional[bool] = Field(False, alias="useLimitFlowPos")
    limitflowpos: Optional[float] = Field(None, alias="limitFlowPos")

    uselimitflowneg: Optional[bool] = Field(False, alias="useLimitFlowNeg")
    limitflowneg: Optional[float] = Field(None, alias="limitFlowNeg")

    @field_validator("allowedflowdir", mode="before")
    @classmethod
    def _flowdirection_validator(cls, v) -> FlowDirection:
        return enum_value_parser(v, FlowDirection)

    @model_validator(mode="after")
    def _validate_limitflowpos(self):
        if self.limitflowpos is None and self.uselimitflowpos:
            raise ValueError(
                "`limitFlowPos` should be defined when `useLimitFlowPos` is True."
            )
        if self.limitflowneg is None and self.uselimitflowneg:
            raise ValueError(
                "`limitFlowNeg` should be defined when `useLimitFlowNeg` is True."
            )
        return self


class GateOpeningHorizontalDirection(StrEnum):
    """Horizontal opening direction of gate door[s]."""

    symmetric = "symmetric"
    from_left = "fromLeft"
    from_right = "fromRight"
    allowedvaluestext = "Possible values: symmetric, fromLeft, fromRight."


class GeneralStructure(Structure):
    """General Structure.

    Hydraulic structure with `type=generalStructure`, to be included in a structure file.
    Typically inside the structure list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.structurefile[0].structure[..]`

    All lowercased attributes match with the orifice input as described in
    [UM Sec.C.12.9](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.12.9).
    """

    class Comments(Structure.Comments):
        type: Optional[str] = Field(
            "Structure type; must read generalStructure", alias="type"
        )
        allowedflowdir: Optional[str] = Field(
            FlowDirection.allowedvaluestext, alias="allowedFlowDir"
        )

        upstream1width: Optional[str] = Field("w_u1 [m]", alias="upstream1Width")
        upstream1level: Optional[str] = Field("z_u1 [m AD]", alias="upstream1Level")
        upstream2width: Optional[str] = Field("w_u2 [m]", alias="upstream2Width")
        upstream2level: Optional[str] = Field("z_u2 [m D]", alias="upstream2Level")

        crestwidth: Optional[str] = Field("w_s [m]", alias="crestWidth")
        crestlevel: Optional[str] = Field("z_s [m AD]", alias="crestLevel")
        crestlength: Optional[str] = Field(
            "The crest length across the general structure [m]. When the crest length > 0, the extra resistance for this structure will be ls * g/(C2 * waterdepth)",
            alias="crestLength",
        )

        downstream1width: Optional[str] = Field("w_d1 [m]", alias="downstream1Width")
        downstream1level: Optional[str] = Field("z_d1 [m AD]", alias="downstream1Level")
        downstream2width: Optional[str] = Field("w_d2 [m]", alias="downstream2Width")
        downstream2level: Optional[str] = Field("z_d2 [m AD]", alias="downstream2Level")

        gateloweredgelevel: Optional[str] = Field(
            "Position of gate door’s lower edge [m AD]", alias="gateLowerEdgeLevel"
        )
        posfreegateflowcoeff: Optional[str] = Field(
            "Positive free gate flow corr.coeff. cgf [-]", alias="posFreeGateFlowCoeff"
        )
        posdrowngateflowcoeff: Optional[str] = Field(
            "Positive drowned gate flow corr.coeff. cgd [-]",
            alias="posDrownGateFlowCoeff",
        )
        posfreeweirflowcoeff: Optional[str] = Field(
            "Positive free weir flow corr.coeff. cwf [-]", alias="posFreeWeirFlowCoeff"
        )
        posdrownweirflowcoeff: Optional[str] = Field(
            "Positive drowned weir flow corr.coeff. cwd [-]",
            alias="posDrownWeirFlowCoeff",
        )
        poscontrcoeffreegate: Optional[str] = Field(
            "Positive gate flow contraction coefficient µgf [-]",
            alias="posContrCoefFreeGate",
        )
        negfreegateflowcoeff: Optional[str] = Field(
            "Negative free gate flow corr.coeff. cgf [-]", alias="negFreeGateFlowCoeff"
        )
        negdrowngateflowcoeff: Optional[str] = Field(
            "Negative drowned gate flow corr.coeff. cgd [-]",
            alias="negDrownGateFlowCoeff",
        )
        negfreeweirflowcoeff: Optional[str] = Field(
            "Negative free weir flow corr.coeff. cwf [-]", alias="negFreeWeirFlowCoeff"
        )
        negdrownweirflowcoeff: Optional[str] = Field(
            "Negative drowned weir flow corr.coeff. cwd [-]",
            alias="negDrownWeirFlowCoeff",
        )
        negcontrcoeffreegate: Optional[str] = Field(
            "Negative gate flow contraction coefficient mu gf [-]",
            alias="negContrCoefFreeGate",
        )
        extraresistance: Optional[str] = Field(
            "Extra resistance [-]", alias="extraResistance"
        )
        gateheight: Optional[str] = Field(None, alias="gateHeight")
        gateopeningwidth: Optional[str] = Field(
            "Opening width between gate doors [m], should be smaller than (or equal to) crestWidth",
            alias="gateOpeningWidth",
        )
        gateopeninghorizontaldirection: Optional[str] = Field(
            "Horizontal opening direction of gate door[s]. Possible values are: symmetric, fromLeft, fromRight",
            alias="gateOpeningHorizontalDirection",
        )
        usevelocityheight: Optional[str] = Field(
            "Flag indicates whether the velocity height is to be calculated or not",
            alias="useVelocityHeight",
        )

    comments: Optional[Comments] = Comments()

    type: Literal["generalStructure"] = Field("generalStructure", alias="type")
    allowedflowdir: Optional[FlowDirection] = Field(
        FlowDirection.both.value, alias="allowedFlowDir"
    )

    upstream1width: Optional[float] = Field(10.0, alias="upstream1Width")
    upstream1level: Optional[float] = Field(0.0, alias="upstream1Level")
    upstream2width: Optional[float] = Field(10.0, alias="upstream2Width")
    upstream2level: Optional[float] = Field(0.0, alias="upstream2Level")

    crestwidth: Optional[float] = Field(10.0, alias="crestWidth")
    crestlevel: Optional[ForcingDataUnion] = Field(0.0, alias="crestLevel")
    crestlength: Optional[float] = Field(0.0, alias="crestLength")

    downstream1width: Optional[float] = Field(10.0, alias="downstream1Width")
    downstream1level: Optional[float] = Field(0.0, alias="downstream1Level")
    downstream2width: Optional[float] = Field(10.0, alias="downstream2Width")
    downstream2level: Optional[float] = Field(0.0, alias="downstream2Level")

    gateloweredgelevel: Optional[ForcingDataUnion] = Field(
        11.0, alias="gateLowerEdgeLevel"
    )
    posfreegateflowcoeff: Optional[float] = Field(1.0, alias="posFreeGateFlowCoeff")
    posdrowngateflowcoeff: Optional[float] = Field(1.0, alias="posDrownGateFlowCoeff")
    posfreeweirflowcoeff: Optional[float] = Field(1.0, alias="posFreeWeirFlowCoeff")
    posdrownweirflowcoeff: Optional[float] = Field(1.0, alias="posDrownWeirFlowCoeff")
    poscontrcoeffreegate: Optional[float] = Field(1.0, alias="posContrCoefFreeGate")
    negfreegateflowcoeff: Optional[float] = Field(1.0, alias="negFreeGateFlowCoeff")
    negdrowngateflowcoeff: Optional[float] = Field(1.0, alias="negDrownGateFlowCoeff")
    negfreeweirflowcoeff: Optional[float] = Field(1.0, alias="negFreeWeirFlowCoeff")
    negdrownweirflowcoeff: Optional[float] = Field(1.0, alias="negDrownWeirFlowCoeff")
    negcontrcoeffreegate: Optional[float] = Field(1.0, alias="negContrCoefFreeGate")
    extraresistance: Optional[float] = Field(0.0, alias="extraResistance")
    gateheight: Optional[float] = Field(1e10, alias="gateHeight")
    gateopeningwidth: Optional[ForcingDataUnion] = Field(0.0, alias="gateOpeningWidth")
    gateopeninghorizontaldirection: Optional[GateOpeningHorizontalDirection] = Field(
        GateOpeningHorizontalDirection.symmetric.value,
        alias="gateOpeningHorizontalDirection",
    )
    usevelocityheight: Optional[bool] = Field(True, alias="useVelocityHeight")


class DambreakAlgorithm(int, Enum):
    """Dambreak algorithm options."""

    van_der_knaap = 1  # "van der Knaap, 2000"
    verheij_van_der_knaap = 2  # "Verheij-van der Knaap, 2002"
    timeseries = 3  # "Predefined time series, dambreakLevelsAndWidths."

    @property
    def description(self) -> str:
        """Description.

        Property to return the description of the enums defined above.
        Useful for comments in output files.

        Returns:
            str: Description for the current enum.
        """
        description_dict = dict(
            van_der_knaap="van der Knaap, 2000",
            verheij_van_der_knaap="Verheij-van der Knaap, 2002",
            timeseries="Predefined time series, dambreakLevelsAndWidths",
        )
        return description_dict[self.name]


class Dambreak(Structure):
    """Dambreak structure.

    Hydraulic structure with `type=dambreak`, to be included in a structure file.
    Typically inside the structure list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.structurefile[0].structure[..]`

    All lowercased attributes match with the dambreak input as described in
    [UM Sec.C.12.10](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.12.10).
    """

    class Comments(Structure.Comments):
        type: Optional[str] = Field("Structure type; must read dambreak", alias="type")
        startlocationx: Optional[str] = Field(
            "x-coordinate of breach starting point.", alias="startLocationX"
        )
        startlocationy: Optional[str] = Field(
            "y-coordinate of breach starting point.", alias="startLocationY"
        )
        algorithm: Optional[str] = Field(
            "Breach growth algorithm. Possible values are: 1 (van der Knaap (2000)), 2 (Verheij–van der Knaap (2002)), 3: Predefined time series, see dambreakLevelsAndWidths",
            alias="algorithm",
        )
        crestlevelini: Optional[str] = Field(
            "Initial breach level zcrest level [m AD].", alias="crestLevelIni"
        )
        breachwidthini: Optional[str] = Field(
            "Initial breach width B0 [m].", alias="breachWidthIni"
        )
        crestlevelmin: Optional[str] = Field(
            "Minimal breach level zmin [m AD].", alias="crestLevelMin"
        )
        t0: Optional[str] = Field("Breach start time Tstart [s].", alias="t0")
        timetobreachtomaximumdepth: Optional[str] = Field(
            "tPhase 1 [s].", alias="timeToBreachToMaximumDepth"
        )
        f1: Optional[str] = Field("Factor f1 [-]", alias="f1")
        f2: Optional[str] = Field("Factor f2 [-]", alias="f2")
        ucrit: Optional[str] = Field(
            "Critical flow velocity uc for erosion [m/s].", alias="uCrit"
        )
        waterlevelupstreamlocationx: Optional[str] = Field(
            "(optional) x-coordinate of custom upstream water level point.",
            alias="waterLevelUpstreamLocationX",
        )
        waterlevelupstreamlocationy: Optional[str] = Field(
            "(optional) y-coordinate of custom upstream water level point.",
            alias="waterLevelUpstreamLocationY",
        )
        waterleveldownstreamlocationx: Optional[str] = Field(
            "(optional) x-coordinate of custom downstream water level point.",
            alias="waterLevelDownstreamLocationX",
        )
        waterleveldownstreamlocationy: Optional[str] = Field(
            "(optional) y-coordinate of custom downstream water level point.",
            alias="waterLevelDownstreamLocationY",
        )
        waterlevelupstreamnodeid: Optional[str] = Field(
            "(optional) Node Id of custom upstream water level point.",
            alias="waterLevelUpstreamNodeId",
        )
        waterleveldownstreamnodeid: Optional[str] = Field(
            "(optional) Node Id of custom downstream water level point.",
            alias="waterLevelDownstreamNodeId",
        )
        dambreaklevelsandwidths: Optional[str] = Field(
            "(only when algorithm=3) Filename of <*.tim> file (Section C.4) containing the breach levels and widths.",
            alias="dambreakLevelsAndWidths",
        )

    comments: Comments = Comments()
    type: Literal["dambreak"] = Field("dambreak", alias="type")
    startlocationx: float = Field(alias="startLocationX")
    startlocationy: float = Field(alias="startLocationY")
    algorithm: DambreakAlgorithm = Field(alias="algorithm")

    crestlevelini: float = Field(alias="crestLevelIni")
    breachwidthini: float = Field(alias="breachWidthIni")
    crestlevelmin: float = Field(alias="crestLevelMin")
    t0: float = Field(alias="t0")
    timetobreachtomaximumdepth: float = Field(alias="timeToBreachToMaximumDepth")
    f1: float = Field(alias="f1")
    f2: float = Field(alias="f2")
    ucrit: float = Field(alias="uCrit")
    waterlevelupstreamlocationx: Optional[float] = Field(
        None, alias="waterLevelUpstreamLocationX"
    )
    waterlevelupstreamlocationy: Optional[float] = Field(
        None, alias="waterLevelUpstreamLocationY"
    )
    waterleveldownstreamlocationx: Optional[float] = Field(
        None, alias="waterLevelDownstreamLocationX"
    )
    waterleveldownstreamlocationy: Optional[float] = Field(
        None, alias="waterLevelDownstreamLocationY"
    )
    waterlevelupstreamnodeid: Optional[str] = Field(
        None, alias="waterLevelUpstreamNodeId"
    )
    waterleveldownstreamnodeid: Optional[str] = Field(
        None, alias="waterLevelDownstreamNodeId"
    )
    dambreaklevelsandwidths: Optional[ForcingDataUnion] = Field(
        None, alias="dambreakLevelsAndWidths"
    )

    @field_validator("algorithm", mode="after")
    @classmethod
    def validate_algorithm(cls, value: str) -> DambreakAlgorithm:
        """
        Validates the algorithm parameter for the dambreak structure.

        Args:
            value (int): algorithm value read from the user's input.

        Raises:
            ValueError: When the value given is not of type int.
            ValueError: When the value given is not in the range [1,3]

        Returns:
            int: Validated value.
        """
        try:
            int_value = int(value)
        except Exception:
            raise ValueError("Dambreak algorithm value should be of type int.")

        if not (0 < int_value <= 3):
            raise ValueError("Dambreak algorithm value should be 1, 2 or 3.")
        return DambreakAlgorithm(int_value)

    @field_validator("dambreaklevelsandwidths", mode="after")
    @classmethod
    def validate_dambreak_levels_and_widths(
        cls, field_value: Optional[Union[TimModel, ForcingModel]], info: ValidationInfo
    ) -> Optional[Union[TimModel, ForcingModel]]:
        """Validate dambreak levels and widths.

        Validates whether a dambreak can be created with the given dambreakLevelsAndWidths
        property. This property should be given when the algorithm value is 3.

        Args:
            field_value (Optional[Union[TimModel, ForcingModel]]): Value given for dambreakLevelsAndWidths.
            values (dict): Dictionary of values already validated (assuming algorithm is in it).

        Raises:
            ValueError: When algorithm value is not 3 and field_value has a value.

        Returns:
            Optional[Union[TimModel, ForcingModel]]: The value given for dambreakLevelsAndwidths.
        """
        # Retrieve the algorithm value (if not found use 0).
        if isinstance(info, dict):
            algorithm_value = info.get("algorithm", 0)
        else:
            algorithm_value = info.data.get("algorithm", 0)
        if field_value is not None and algorithm_value != 3:
            # dambreakLevelsAndWidths can only be set when algorithm = 3
            raise ValueError(
                f"Dambreak field dambreakLevelsAndWidths can only be set when algorithm = 3, current value: {algorithm_value}."
            )
        return field_value

    @model_validator(mode="after")
    def check_location_dambreak(self) -> "Dambreak":
        """Check location dambreak.

            - Verifies whether the location for this structure contains valid values for numCoordinates, xCoordinates and
                yCoordinates or instead is using a polyline file.
            - Verifies whether de water level location specifications are valid.

        Raises:
            ValueError: When the values dictionary does not contain valid coordinates or polyline file or when the water level location specifications are not valid.

        Returns:
            dict: Dictionary of validated values.
        """
        self._validate_waterlevel_location(
            self.model_dump(),
            "waterLevelUpstreamLocationX",
            "waterLevelUpstreamLocationY",
            "waterLevelUpstreamNodeId",
        )
        self._validate_waterlevel_location(
            self.model_dump(),
            "waterLevelDownstreamLocationX",
            "waterLevelDownstreamLocationY",
            "waterLevelDownstreamNodeId",
        )

        return self

    @staticmethod
    def _validate_waterlevel_location(values, x_key: str, y_key: str, node_key: str):
        x_is_given = values.get(x_key.lower()) is not None
        y_is_given = values.get(y_key.lower()) is not None
        node_is_given = values.get(node_key.lower()) is not None

        if not (
            (x_is_given and y_is_given and not node_is_given)
            or (node_is_given and not x_is_given and not y_is_given)
        ):
            raise ValueError(
                f"Either `{node_key}` should be specified or `{x_key}` and `{y_key}`."
            )


class Bridge(Structure):
    """Bridge structure.

    Hydraulic structure with `type=bridge`, to be included in a structure file.
    Typically inside the structure list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.structurefile[0].structure[..]`

    All lowercased attributes match with the bridge input as described in
    [UM Sec.C.12.5](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.C.12.5).
    """

    class Comments(Structure.Comments):
        type: Optional[str] = Field("Structure type; must read bridge", alias="type")
        allowedflowdir: Optional[str] = Field(
            FlowDirection.allowedvaluestext, alias="allowedFlowdir"
        )

        csdefid: Optional[str] = Field(
            "Id of Cross-Section Definition.", alias="csDefId"
        )
        shift: Optional[str] = Field(
            "Vertical shift of the cross section definition [m]. Defined positive upwards."
        )
        inletlosscoeff: Optional[str] = Field(
            "Inlet loss coefficient [-], ξ_i.",
            alias="inletLossCoeff",
        )
        outletlosscoeff: Optional[str] = Field(
            "Outlet loss coefficient [-], k.",
            alias="outletLossCoeff",
        )
        frictiontype: Optional[str] = Field(
            "Friction type, possible values are: Chezy, Manning, wallLawNikuradse, WhiteColebrook, StricklerNikuradse, Strickler, deBosBijkerk.",
            alias="frictionType",
        )
        friction: Optional[str] = Field(
            "Friction value, used in friction loss.",
            alias="friction",
        )
        length: Optional[str] = Field("Length [m], L.")

    comments: Comments = Comments()

    type: Literal["bridge"] = Field("bridge", alias="type")
    allowedflowdir: FlowDirection = Field(alias="allowedFlowdir")

    csdefid: str = Field(alias="csDefId")
    shift: float
    inletlosscoeff: float = Field(alias="inletLossCoeff")
    outletlosscoeff: float = Field(alias="outletLossCoeff")
    frictiontype: FrictionType = Field(alias="frictionType")
    friction: float
    length: float

    @field_validator("frictiontype", mode="before")
    @classmethod
    def _frictiontype_validator(cls, v) -> FrictionType:
        return enum_value_parser(v, FrictionType)


class StructureGeneral(INIGeneral):
    """`[General]` section with structure file metadata."""

    _header: Literal["General"] = "General"
    fileversion: str = Field("3.00", alias="fileVersion")
    filetype: Literal["structure"] = Field("structure", alias="fileType")


StructureUnion = Annotated[
    Union[
        Weir,
        UniversalWeir,
        Culvert,
        LongCulvert,
        Pump,
        Compound,
        Orifice,
        GeneralStructure,
        Dambreak,
        Bridge,
    ],
    Field(discriminator="type"),
]


class StructureModel(INIModel):
    """
    The overall structure model that contains the contents of one structure file.

    This model is typically referenced under a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.structurefile[..]`.

    Attributes:
        general (StructureGeneral):
            `[General]` block with file metadata.
        structure (List[Structure]):
            List of `[Structure]` blocks for all hydraulic structures.
    """

    general: StructureGeneral = StructureGeneral()
    structure: Annotated[List[StructureUnion], BeforeValidator(make_list)] = []

    @classmethod
    def _ext(cls) -> str:
        return ".ini"

    @classmethod
    def _filename(cls) -> str:
        return "structures"



"""MDU Parser."""

from collections import Counter
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

from hydrolib.core.base.file_manager import PathOrStr
from hydrolib.core.dflowfm.mdu.models import FMModel, Physics, Time
from hydrolib.tools.extforce_convert.utils import (
    CONVERTER_DATA,
    IgnoreUnknownKeyWordClass,
    backup_file,
)

STRUCTURE_FILE_LINE = "StructureFile"
INIFIELD_FILE_LINE = "IniFieldFile"
PATH_RELATIVE_TO_PARENT = "PathsRelativeToParent"

__all__ = ["MDUParser"]


@dataclass
class Section:
    """Information about a section in an MDU file.

    Attributes:
        start (int):
            The index of the section header line (0-index).
        end (int):
            The index of the last line in the section (0-index), the last line in the section is the line before the
            header of the next section.
        non_key_value_lines_at_end (int):
            number of empty or non-key-value lines at the end of the section.
        last_key_value_line_index (int):
            index of the last key-value line in the section (0-based index), use this index to append any key-value
            to the end of the section.
    Notes:
        - The section header is expected to be in the format "[section_name]".
        - The section is defined as starting from the section header line and ending at the line before the header of
        the next section.
        - If all the lines in the section are empty lines or comments, the `start` and `last_key_value_line_index`
        will be the same.
        - if the section is not found, `start` and `end` will be set to None.
        - if the section exists twice in the file, the first occurrence will be used.
    """

    start: Optional[int]
    end: Optional[int]
    non_key_value_lines_at_end: Optional[int]
    last_key_value_line_index: Optional[int]

    def __init__(self, section_name: str, content: List[str]):
        """Initialize SectionBounds with content and section indices."""
        section_start = -1
        section_end = len(content) - 1

        header = f"[{section_name.lower()}]"

        # Find section start and end
        for i, line in enumerate(content):
            stripped = line.strip().lower()
            if stripped.startswith("[") and stripped.endswith("]"):
                if stripped == header:
                    section_start = i
                elif section_start != -1:
                    # We hit the start of a new section after finding our target
                    section_end = i - 1
                    break

        if section_start == -1:
            # Section not found
            section_start = None
            section_end = None

        self.start = section_start
        self.end = section_end
        self.non_key_value_lines_at_end = (
            None
            if section_start is None
            else (
                self.get_empty_and_non_key_value_lines(
                    content, section_start, section_end
                )
            )
        )
        self.last_key_value_line_index = (
            None
            if section_start is None
            else (self.end - self.non_key_value_lines_at_end)
        )

    @staticmethod
    def get_empty_and_non_key_value_lines(
        content, section_start: int, section_end: int
    ):
        # Find empty lines at the end of the section and non-key-value lines within the section
        empty_lines_at_end = []

        if section_start is not None and section_end is not None:
            # Check for empty lines at the end of the section
            for i in range(section_end, section_start, -1):
                line_obj = Line(content[i])
                if line_obj.is_empty() or line_obj.is_comment():
                    empty_lines_at_end.append(i)
                else:
                    break

        return len(empty_lines_at_end)


@dataclass
class ExternalForcingBlock:
    extforcefile: Union[Path, str]
    extforcefilenew: Optional[Union[Path, str]] = field(default=None)
    comments: Optional[List[str]] = field(default=None)
    _header: Optional = "[external forcing]"
    root_dir: Optional[Path] = field(default=None)

    def __init__(self, **kwargs):
        valid_keys = {"extforcefile", "extforcefilenew", "comments", "_header"}
        for key in valid_keys:
            setattr(self, key, kwargs.get(key, None))

    @property
    def extforce_file(self) -> Path:
        old_ext_force_file = self.extforcefile
        if old_ext_force_file is None:
            raise ValueError(
                "An old formatted external forcing file (.ext) could not be found in the mdu file.\n"
                "Conversion is not possible or may not be necessary."
            )
        else:
            old_ext_force_file = Path(old_ext_force_file)

        return old_ext_force_file

    def get_new_extforce_file(self, ext_file: Optional[Path] = None) -> Path:
        """Get the new external forcing file path.

        Notes:
            - If the `extforcefilenew` exists in the MDU file, it will be used.
            - If it does not exist, it will create a new file with the old extforce file name with a "-new" suffix.
            - If an `ext_file` is provided, it will be used as the new extforce file.

        Args:
            ext_file (Path):
                Optional path to an external forcing file to use as the new extforce file.

        Returns:
            Path:
                Path to the new external forcing file.
        """
        _extforce_file_new = (
            Path(self.extforcefilenew) if self.extforcefilenew else None
        )

        if _extforce_file_new:
            # if the extforce_file_new exist in the MDU file, we use it
            ext_file = (self.root_dir / _extforce_file_new).resolve()
        else:
            # if the extforce_file_new does not exist in the MDU file
            if ext_file is None:
                # if no ext_file is provided, we use the old extforce file name to create the new extforce file
                ext_file = self.root_dir / self.extforce_file.with_stem(
                    self.extforce_file.stem + "-new"
                )
                if ext_file.exists():
                    raise FileExistsError(
                        "The converter detected that there is no new extforce file in the mdu file, \n"
                        f"But there is an extforce file with the name {ext_file} that already exists. \n"
                        "Please either remove/rename the file or add it to the mdu file. in the section [external forcing]"
                    )
            else:
                # if an ext_file is provided, we use it
                ext_file = Path(ext_file).resolve()

        return ext_file


@dataclass
class FileStyleProperties:
    """
    Detects and stores style properties of an MDU file, such as leading spaces and equal sign alignment.

    This class analyzes the content of an MDU file to determine the most common number of leading spaces
    and the most common position of the equal sign for key-value pairs. These properties can be used to
    preserve formatting when updating or generating MDU files.

    Attributes:
        leading_spaces (int): The most common number of leading spaces at the start of lines.
        equal_sign_position (int): The most common position (column index) of the equal sign in key-value lines.

    Example:
        ```python
        >>> content = [
        ...     'Param1    = value1',
        ...     'Param2    = value2',
        ...     'Param3    = value3',
        ... ]
        >>> style = FileStyleProperties(content)
        >>> style.leading_spaces
        0
        >>> style.equal_sign_position
        9
        ```
    """

    leading_spaces: int = 0
    equal_sign_position: int = 0
    comments_position: Optional[int] = None

    def __init__(self, content: List[str]):
        """
        Initialize the FileStyleProperties.

        Args:
            content (List[str]): List of strings representing the content of the MDU file.
        """
        self.leading_spaces = self._get_leading_spaces(content)
        self.equal_sign_position = self._get_equal_sign_position(content)
        self.comments_position = self._get_comments_position(content)

    @staticmethod
    def _get_equal_sign_position(content: List[str]) -> Optional[int]:
        """
        Get the most common position (column index) of the equal sign in the MDU file.

        Args:
            content (List[str]): List of strings representing the content of the MDU file.

        Returns:
            Optional[int]: The most common index of the equal sign, or None if not found.

        Example:
            ```python
            >>> FileStyleProperties._get_equal_sign_position(
            ...     ['A = 1', 'BB   = 2', 'CCC = 3', '# Comment = 4']
            ... )
            2
            ```
        """
        position = FileStyleProperties._locate(content, "=")

        return position

    def _get_comments_position(self, content: List[str]) -> Optional[int]:
        """
        Get the most common position (column index) of the comment character in the MDU file.

        Args:
            content (List[str]):
                List of strings representing the content of the MDU file.

        Returns:
            Optional[int]:
                The most common index of the comment character, or None if not found.

        Notes:
            - This function is used to determine where comments start in the MDU file.
            - It ignores lines that start with the comment character `#` to avoid counting comments as part of the key-value pairs.
            - The function uses a Counter to find the most common position of the comment character.
            - If no lines with comments are found, it returns None.
            - if the line has multiple comment characters, it will only count the first one.
        """
        position = self._locate(content, "#")

        return position

    @staticmethod
    def _locate(content: List[str], keyword: str) -> Optional[int]:
        """
        Get the most common position (column index) of the equal sign in the MDU file.

        Args:
            content (List[str]): List of strings representing the content of the MDU file.

        Returns:
            Optional[int]: The most common index of the equal sign, or None if not found.

        Example:
            ```python
            >>> FileStyleProperties._get_equal_sign_position(
            ...     ['A = 1', 'BB   = 2', 'CCC = 3', '# Comment = 4']
            ... )
            2
            ```
        """
        equal_sign_counter = Counter()
        for line in content:
            if "=" in line and not line.strip().startswith("#"):
                eq_index = line.find(keyword)
                if eq_index != -1:  # when the keyword is not found, eq_index will be -1
                    equal_sign_counter[eq_index] += 1

        position = None
        if equal_sign_counter:
            most_common_pos, _ = equal_sign_counter.most_common(1)[0]
            position = most_common_pos

        return position

    @staticmethod
    def _get_leading_spaces(content: List[str]) -> Optional[int]:
        """Get Leading Spaces.

        - Detect the most common number of leading spaces in each line of the MDU file.
        - The function returns the most common leading space count, which can be used to maintain consistent formatting

        Args:
            content (List[str]):
                List of strings representing the content of the MDU file.

        Returns:
            Optional[int]:
                The most common number of leading spaces, or None if not found.

        Notes:
            - All the lines that start with a comment character `#` are ignored.
            - The leading spaces are counted as the difference between the length of the line and the length of the line
                after stripping leading whitespace.
            - If no lines with leading spaces are found, returns None.
            - The function uses a Counter to find the most common leading space count.
            - tabs and spaces are treated the same, so if a line has 1 spaces and 1 tab, it will be counted as 2
            leading spaces.

        Example:
            ```python
            >>> FileStyleProperties._get_leading_spaces([
            ...     '  Param1 = value1', '    Param2 = value2', '  Param3 = value3'])
            2
            ```
        """
        leading_spaces = Counter()
        for line in content:
            if not line.strip().startswith("#"):
                num_spaces = len(line) - len(line.lstrip())
                leading_spaces[num_spaces] += 1

        spacing = None
        if leading_spaces:
            most_common_spacing, _ = leading_spaces.most_common(1)[0]
            spacing = most_common_spacing

        return spacing


class Line:

    def __init__(self, line: str):
        self._content = line
        self.validate()

    def __str__(self):
        """String representation of the Line object."""
        return self.content

    @property
    def content(self) -> str:
        """Get the content of the line."""
        return self._content

    def validate(self):
        """Validate the line content."""
        if not (self.is_empty() or self.is_comment() or self.is_section_header()):
            if self.equal_sign_position is None:
                raise ValueError(
                    f"Invalid line content: '{self._content}'. Line must contain an equal sign '=' for key-value pairs."
                )

    @property
    def key_value(self) -> Tuple[str, str]:
        """Get the key of the line."""
        if self.is_comment() or self.is_section_header():
            key = None
            value = None
        else:
            key, value = self.get_key_value()

        return key, value

    @property
    def key(self) -> str:
        """Get the key of the line."""
        key, _ = self.key_value
        return key

    @property
    def value(self) -> str:
        """Get the value of the line."""
        _, value = self.key_value
        return value

    @property
    def equal_sign_position(self) -> int:
        """Get the position of the equal sign in the line."""
        return self.content.find("=") if "=" in self.content else None

    @property
    def comments(self) -> str:
        _, comment = self._get_comments()
        return comment

    @property
    def comment_position(self) -> Optional[int]:
        """Get the position of the comments in the line."""
        comment_index = self.content.find("#")
        if comment_index == -1:
            comment_index = None

        return comment_index

    @property
    def leading_spaces(self) -> int:
        """Get the number of leading spaces in the line."""
        return self.get_leading_spaces()

    def _get_comments(self) -> Tuple[int, str]:
        """Get comments from the line."""
        comment_index = self.content.find("#")
        if comment_index != -1:
            comments = self.content[comment_index:].strip()
            comments += "\n"
        else:
            comment_index = None
            comments = ""

        return comment_index, comments

    def is_comment(self) -> bool:
        """Check if the line is a comment."""
        return self.content.strip().startswith("#")

    def is_section_header(self) -> bool:
        """Check if the line is a section header (e.g., '[...]')."""
        return self.content.startswith("[") and self.content.endswith("]")

    def is_empty(self) -> bool:
        """Check if the line is empty or contains only whitespace."""
        return not self.content.strip()

    def get_key_value(self) -> Tuple[str, str]:
        """Get the key and value from the line."""
        if self.is_comment() or self.is_section_header() or self.is_empty():
            key = None
            value = None
        else:
            if "=" in self.content:
                key, value = self.content.split("=", 1)
                key = key.strip()
                value = value.strip()
                if "#" in value:
                    value = value.split("#", 1)[0].strip()
                    if value == "":
                        value = None
            else:
                raise ValueError(f"Line '{self.content}' does not contain an '='")

        return key, value

    def get_leading_spaces(self) -> int:
        """Get the number of leading spaces in the line."""
        return len(self.content) - len(self.content.lstrip())

    def recenter_equal_sign(
        self, equal_sign_position: int = None, leading_spaces: int = None
    ) -> str:
        """Recenter Equal Sign.

        Recenter the equal sign to a specific target column.

        Args:
            equal_sign_position (int, optional):
                location of the equal sign in the line, if None, it will use the position of the equal sign in the line.
            leading_spaces (int, optional):
                number of leading spaces in the line, if None, it will use the number of leading spaces in the line.

        Returns:
            str:
                Re-aligned line with equal sign at target_pos

        Notes:
            - If the line does not contain an equal sign, it will raise a ValueError.
            - The function preserves the leading spaces and aligns the key to the specified equal sign position.
        """
        key, _ = self.get_key_value()

        if equal_sign_position is None:
            equal_sign_position = self.equal_sign_position

        if leading_spaces is None:
            leading_spaces = self.leading_spaces

        aligned_key = key.ljust(equal_sign_position - leading_spaces)
        spaces = " " * leading_spaces
        all_after_equal_sign = self._content[self.equal_sign_position + 1 :]
        self._content = f"{spaces}{aligned_key}= {all_after_equal_sign.lstrip()}"

    def recenter_comments(self, comments_position: int = None) -> str:
        """Recenter Comments.

        Recenter the comments to a specific target column.

        Args:
            comments_position (int, optional):
                position of the comments in the line, if None, it will use the position of the comments in the line.

        Returns:
            str:
                Re-aligned line with comments at target_pos
        """
        if comments_position is None:
            comments_position = self.comment_position

        if self.comment_position is not None:
            aligned = self.content[: self.comment_position].ljust(comments_position)
            self._content = f"{aligned}{self.comments}"

    @classmethod
    def from_key_value(
        cls,
        key: str,
        value: str,
        equal_sign_position: int = None,
        leading_spaces: int = 0,
        comment: str = None,
    ) -> "Line":
        """Create a Line instance from a key-value pair.

        Args:
            key (str): The key to set in the line. Must not be empty.
            value (str): The value to set in the line. Can be empty.
            equal_sign_position (int, optional): Position of the equal sign. Defaults to None.
            leading_spaces (int, optional): Number of leading spaces. Defaults to 0.
            comment (str, optional): Comment to append to the line. Defaults to None.

        Returns:
            Line: A new Line instance with the specified key and value.

        Raises:
            ValueError: If key is not a string or is empty, or if positions are negative.

        Examples:
            >>> Line.from_key_value('Param', 'value')
            Line('Param= value')
            >>> Line.from_key_value('Param', '', comment='# no value')
            Line('Param=  # no value')
        """
        if not isinstance(key, str):
            raise ValueError("Key and value must be strings.")
        if key == "":
            raise ValueError("Key must not be empty.")
        if equal_sign_position is not None and equal_sign_position < 0:
            raise ValueError("equal_sign_position cannot be negative.")
        if leading_spaces < 0:
            raise ValueError("leading_spaces cannot be negative.")
        if equal_sign_position is None:
            equal_sign_position = len(key) + leading_spaces + 1

        aligned_key = key.ljust(equal_sign_position - leading_spaces)
        spaces = " " * leading_spaces
        line_content = (
            f"{spaces}{aligned_key}= {value}"
            if value != ""
            else f"{spaces}{aligned_key}= "
        )

        if comment:
            stripped = comment.strip()
            if not stripped.endswith("\n"):
                stripped += "\n"
            line_content = f"{line_content} {stripped}"
        else:
            line_content += "\n"

        return cls(line_content)

    def update_value(self, new_value: str):
        """Update the value of the line, preserving key and formatting."""
        if self.is_comment() or self.is_section_header() or self.is_empty():
            raise ValueError("Cannot update value for comment or section header lines.")

        key, _ = self.get_key_value()

        # Rebuild the line
        return Line.from_key_value(
            key,
            new_value,
            equal_sign_position=self.equal_sign_position,
            leading_spaces=self.leading_spaces,
            comment=self.comments,
        )


class MDUParser:
    """A class to update the ExtForceFileNew entry in an MDU file."""

    def __init__(self, mdu_path: PathOrStr):
        """Initialize the MDUParser.

        Args:
            mdu_path: Path to the MDU file to update
        """
        mdu_path = Path(mdu_path) if not isinstance(mdu_path, Path) else mdu_path
        if not mdu_path.exists():
            raise FileNotFoundError(f"File not found: {mdu_path}")

        self.mdu_path = mdu_path
        self._content = self._read_file()
        self.loaded_fm_data = self._load_with_fm_model()
        self.extforce_block = ExternalForcingBlock(**self.external_forcing)
        self.extforce_block.root_dir = self.mdu_path.parent
        self.temperature_salinity_data = self.get_temperature_salinity_data()
        self._geometry = self.loaded_fm_data.get("geometry")
        self.file_style_properties = FileStyleProperties(self._content)

    def __repr__(self):
        message = f"""
            path: {self.mdu_path}
            No lines: {len(self._content)}
            inifield: {self.has_inifield_file()}
            structure: {self.has_structure_file()} 
        """
        return message

    def _load_with_fm_model(self) -> Dict[str, Any]:
        """Load the MDU file using the FMModel class.

        Returns:
            data (Dict[str, str]):
                all the data inside the mdu file, with each section in the file as a key and the data of that section is
                the value of that key.
        """
        data = FMModel._load(FMModel, self.mdu_path)
        return data

    @property
    def external_forcing(self) -> Dict[str, Any]:
        return self.loaded_fm_data.get("external_forcing")

    @property
    def geometry(self) -> Dict[str, Any]:
        """Get the geometry data from the MDU file."""
        return self._geometry

    @property
    def new_forcing_file(self) -> Union[Path, str]:
        """Get the filename for the ExtForceFileNew entry."""
        if not hasattr(self, "_new_forcing_file"):
            raise AttributeError("new_forcing_file not set")
        return self._new_forcing_file

    @new_forcing_file.setter
    def new_forcing_file(self, file_name: Union[Path, str]) -> None:
        """Set the new filename for the ExtForceFileNew entry.

        Args:
            file_name(PathOrStr):
                New filename for the ExtForceFileNew entry
        """
        if not isinstance(file_name, (Path, str)):
            raise ValueError("new_forcing_filename must be a str or Path")
        self._new_forcing_file = file_name

    @property
    def is_relative_to_parent(self) -> bool:
        """Check if the `PathsRelativeToParent` entry is set to 1.

        - True (1) if the file paths are relative to the direct parent file.
        - False (0) if the file paths are relative to mdu file.
        - Default is False, relative to the mdu file.
        """
        val = self.get_keyword(PATH_RELATIVE_TO_PARENT)
        if val is None:
            val = False
        else:
            if val == "1":
                val = True
            else:
                val = False

        return val

    def get_keyword(self, keyword: str) -> Optional[str]:
        """Get the value of the specified keyword.

        Args:
            keyword (str):
                the keyword to get the value of.

        Returns:
            str:
                The value of the specified keyword if exists, None if does not exist.

        Notes:
            - The method is not a case-sensitive.
            - If the keyword is not found, it will return None.
            - If the is found, but the value is empty, it will return "" (empty string after =).
        """
        ind = self.find_keyword_lines(keyword)

        if ind is None:
            val = None
        else:
            line = Line(self.content[ind])
            val = line.value
        return val

    def _read_file(self) -> List[str]:
        """Read the MDU file into a list of strings.

        Returns:
            List of strings, one for each line in the file
        """
        with open(self.mdu_path, "r", encoding="utf-8") as f:
            lines = f.readlines()
        return lines

    def save(self, backup: bool = False) -> None:
        """Save the updated MDU file."""
        if backup:
            backup_file(self.mdu_path)

        with open(self.mdu_path, "w", encoding="utf-8") as file:
            file.writelines(self.content)

    @property
    def content(self) -> List[str]:
        """Get the content of the MDU file.

        Returns:
            List of strings, one for each line in the file
        """
        return self._content

    @content.setter
    def content(self, new_content: List[str]) -> None:
        """Set the content of the MDU file.

        Args:
            new_content: New content for the MDU file
        """
        self._content = new_content

    def has_inifield_file(self) -> bool:
        """Check if the MDU file has an inifield file defined.

        Returns:
            bool: True if an inifield file is defined, False otherwise
        """
        return self.has_field(INIFIELD_FILE_LINE)

    def has_structure_file(self) -> bool:
        """Check if the MDU file has an structure file defined.

        Returns:
            bool: True if an structure file is defined, False otherwise
        """
        return self.has_field(STRUCTURE_FILE_LINE)

    def has_field(self, field_name: str) -> bool:
        """Check if the MDU file has a given file defined.

        Returns:
            bool:
                True if an inifield file is defined, False otherwise
        """
        return True if self.find_keyword_lines(field_name) is not None else False

    def update_file_entry(
        self, field_name: str, file_name: str, section_name: str
    ) -> None:
        leading_spaces = self.file_style_properties.leading_spaces
        equal_sign_position = self.file_style_properties.equal_sign_position

        if not self.has_field(field_name):
            # if the field does not exist, we create a new line for it and add it to the end of the section
            line = Line.from_key_value(
                field_name,
                file_name,
                leading_spaces=leading_spaces,
                equal_sign_position=equal_sign_position,
            )
            section = self.get_section(section_name)
            # put the inifield file at the end of the geometry section
            line_number = section.last_key_value_line_index + 1
            self.insert_line(line.content, line_number)
        else:
            # if the field already exists, we update it
            # find the line number of the existing field
            existing_field_line_num = self.find_keyword_lines(field_name)
            line = Line(self.content[existing_field_line_num])
            # only update the value if there was no value.
            if not line.value:
                line = line.update_value(file_name)
                line.recenter_comments(self.file_style_properties.comments_position)
                line.recenter_equal_sign(
                    equal_sign_position=equal_sign_position,
                    leading_spaces=leading_spaces,
                )

                if existing_field_line_num is not None:
                    # remove the old line
                    self.content.pop(existing_field_line_num)

                line_number = existing_field_line_num

                self.insert_line(line.content, line_number)

    def update_inifield_file(self, file_name: str) -> None:
        """Update the IniFieldFile entry in the MDU file.

        Args:
            file_name (str):
                The path to the new inifield file to set

        Notes:
            - The method adds the IniFieldFile entry at the end of the geometry section.
            - If the IniFieldFile entry exists with a non-empty value, it is left unchanged
            if it exists but has no value, it is populated with the new value.
        """
        self.update_file_entry(INIFIELD_FILE_LINE, file_name, "geometry")

    def update_structure_file(self, file_name: str) -> None:
        """Update the IniFieldFile entry in the MDU file.

        Args:
            file_name (str):
                The path to the new inifield file to set

        Notes:
            - The method adds the IniFieldFile entry at the end-1 of the geometry section, as some mdu files has
            decorative lines (i.e. "#=========") around the section headers, and the function that detects the end of
            the section detects the end of the section by looking for the next section header. and then this
            decorative line will be considered as the last line in the section and adding the inifield file at
            end_ind - 1 will leave an empty line between the actual last line in the section and the newely added
            inifield file line.
        """
        self.update_file_entry(STRUCTURE_FILE_LINE, file_name, "geometry")

    def update_extforce_file_new(
        self, file_name: str, num_quantities: int, remove_old_ext_file: bool = False
    ) -> None:
        if num_quantities > 0:
            self.update_file_entry("ExtForceFileNew", file_name, "external forcing")
        else:
            ext_force_line = self.find_keyword_lines("ExtForceFileNew")
            if ext_force_line is not None:
                self.content.pop(ext_force_line)

        if remove_old_ext_file:
            old_ext_force_line = self.find_keyword_lines("ExtForceFile")
            if old_ext_force_line is not None:
                self.content.pop(old_ext_force_line)

    def get_temperature_salinity_data(self) -> Dict[str, Any]:
        """Get the info needed from the mdu to process and convert the old external forcing files.

        Returns:
            temperature_and_salinity_info (Dict[str, str]):
                dictionary with the information needed for the conversion tool to convert the `SourceSink` and
                `Boundary` quantities. The dictionary will have three keys `temperature`, `salinity`, and `refdate`.

        Examples:
            ```python
            >>> mdu_info = { #doctest: +SKIP
            ...     "refdate": "minutes since 2015-01-01 00:00:00",
            ...     "temperature": True,
            ...     "salinity": True,
            ... }
            ```
        """
        # read sections of the mdu file.
        time_data = self.loaded_fm_data.get("time")
        physics_data = self.loaded_fm_data.get("physics")
        mdu_time = IgnoreUnknownKeyWordClass(Time, **time_data)
        mdu_physics = IgnoreUnknownKeyWordClass(Physics, **physics_data)

        ref_time = get_ref_time(mdu_time.refdate)
        temperature_and_salinity_info = {
            "file_path": self.mdu_path,
            "refdate": ref_time,
            "temperature": False if mdu_physics.temperature == 0 else True,
            "salinity": mdu_physics.salinity,
        }
        return temperature_and_salinity_info

    def find_keyword_lines(
        self, keyword: str, case_sensitive: bool = False
    ) -> Union[int, None]:
        """Find line numbers in the MDU file where the keyword appears.

        Args:
            keyword: The keyword to search for.
            case_sensitive: Whether the search should be case-sensitive.

        Returns:
            A list of line number where the keyword is found.
        """
        if not case_sensitive:
            keyword = keyword.lower()
        line_number = None
        for i, line in enumerate(self.content, start=0):
            haystack = line if case_sensitive else line.lower()
            stripped_line = haystack.lstrip()
            if_exist = (
                stripped_line.startswith(keyword)
                if case_sensitive
                else stripped_line.lower().startswith(keyword.lower())
            )
            if if_exist:
                line_number = i
                break

        return line_number

    def insert_line(self, line: str, index: int) -> None:
        """Insert a line at the specified index in the MDU file content.

        Args:
            index: The 0-based index where the line should be inserted.
            line: The line to insert (a newline `\n` will be added if not present).

        Raises:
            IndexError: If the index is out of bounds.
        """
        if not line.endswith("\n"):
            line += "\n"

        if index < 0 or index > len(self.content):
            raise IndexError("Index out of bounds for inserting line.")

        self.content.insert(index, line)

    def delete_line(
        self,
        *,
        index: Optional[int] = None,
        keyword: Optional[str] = None,
        case_sensitive: bool = False,
    ) -> None:
        """
        Delete a line from the MDU file content by index or keyword.

        Exactly one of `index` or `keyword` must be provided. If `index` is given, deletes the line at that index.
        If `keyword` is given, deletes the first line starting with the keyword (case sensitivity optional).

        Args:
            index (Optional[int]):
                The 0-based index of the line to delete. Must be within bounds of the file content.
            keyword (Optional[str]):
                The keyword to search for. Deletes the first line starting with this keyword. Cannot be empty string.
                if the keyword does not exist in the file the function will not raise an error.
            case_sensitive (bool):
                Whether the keyword search should be case-sensitive. Defaults to False.

        Returns:
            None

        Raises:
            ValueError: If both or neither of 'index' and 'keyword' are provided, or if keyword is empty string.
            IndexError: If 'index' is out of bounds for the file content.

        Examples:
            - Delete a line by index:
                ```python
                >>> from hydrolib.tools.extforce_convert.mdu_parser import MDUParser
                >>> path = "tests/data/input/mdu-files/extforce_test_1.mdu"
                >>> mdu_file = MDUParser(path)
                >>> mdu_file.content[:2]  # doctest: +SKIP
                ['# Generated by the Matlab program dflowfmConverter on 14:39:00, 27-07-2017\n',
                 '# Deltares, Matlab-generated mdu-file for D-Flow FM\n']
                >>> mdu_file.delete_line(index=0)  # doctest: +SKIP
                >>> mdu_file.content[:2]  # doctest: +SKIP
                ['# Deltares, Matlab-generated mdu-file for D-Flow FM\n', '\n']

                ```

            - Delete a line by keyword (case-insensitive):
                ```python
                >>> path = "tests/data/input/mdu-files/extforce_test_1.mdu"
                >>> mdu_file = MDUParser(path)
                >>> mdu_file.content[8:11]
                ['[geometry]\n',
                 '    NetFile                             = sp_net.nc                          # *_net.nc\n',
                 '    unknown_geometry                    = 5                                   # any comment\n']
                >>> mdu_file.delete_line(keyword='NetFile')
                >>> mdu_file.content[8:11]
                ['[geometry]\n',
                 '    unknown_geometry                    = 5                                   # any comment\n',
                 '    WaterLevIniFile                     =                                    # Initial water levels sample file *.xyz\n']

                ```

            - Delete a line by keyword (case-sensitive) - the function will not find the keyword and will not raise
                an error:
                ```python
                >>> path = "tests/data/input/mdu-files/extforce_test_1.mdu"
                >>> mdu_file = MDUParser(path)
                >>> mdu_file.content[8:11]
                ['[geometry]\n',
                 '    NetFile                             = sp_net.nc                          # *_net.nc\n',
                 '    unknown_geometry                    = 5                                   # any comment\n']
                >>> mdu_file.delete_line(keyword='netfile', case_sensitive=True)
                >>> mdu_file.content[8:11]
                ['[geometry]\n',
                 '    NetFile                             = sp_net.nc                          # *_net.nc\n',
                 '    unknown_geometry                    = 5                                   # any comment\n']
                ```

            - Error if both index and keyword are provided:
                ```python
                >>> path = "tests/data/input/mdu-files/extforce_test_1.mdu"
                >>> mdu_file = MDUParser(path)
                >>> mdu_file.delete_line(index=1, keyword='B') #doctest: +SKIP
                Traceback (most recent call last):
                ...
                ValueError: Exactly one of 'index' or 'keyword' must be provided
                ```

            - Error if index is out of bounds:

                ```python
                >>> from hydrolib.tools.extforce_convert.mdu_parser import MDUParser
                >>> path = "tests/data/input/mdu-files/extforce_test_1.mdu"
                >>> mdu_file = MDUParser(path)
                >>> mdu_file.delete_line(index=1000)
                Traceback (most recent call last):
                ...
                IndexError: index out of range: 1000
                ```

            - Error if keyword is empty string:

                ```python
                >>> from hydrolib.tools.extforce_convert.mdu_parser import MDUParser
                >>> path = "tests/data/input/mdu-files/extforce_test_1.mdu"
                >>> mdu_file = MDUParser(path)
                >>> mdu_file.delete_line(keyword='')
                Traceback (most recent call last):
                ...
                ValueError: keyword cannot be empty string
                ```

        See Also:
            MDUParser.insert_line: For inserting lines into the MDU file content.
            MDUParser.find_keyword_lines: For finding line indices by keyword.
        """
        if (index is None) == (keyword is None):
            raise ValueError("Exactly one of 'index' or 'keyword' must be provided")

        if index is not None:
            # validate bounds explicitly
            if not (0 <= index < len(self.content)):
                raise IndexError(f"index out of range: {index}")
            self.content.pop(index)
        elif keyword is not None:
            if keyword == "":
                raise ValueError("keyword cannot be empty string")

            index = self.find_keyword_lines(keyword, case_sensitive=case_sensitive)

            if index is not None:
                self.content.pop(index)

    def clean(self):
        """
        Remove the deprecated mdu keywords from the file

        Notes:
            - The `clean` function only removes the first occurrence of the deprecated keywords from the file. if a
            keyword is repeated in the file, the `clean` function will only remove the first one.
            not remove the deprecated
        """

        for keyword in CONVERTER_DATA.mdu.deprecated_keywords:
            ind = self.find_keyword_lines(keyword)
            if ind is not None:
                line = Line(self.content[ind])
                if line.value == str(CONVERTER_DATA.mdu.deprecated_value):
                    self.content.pop(ind)

    def get_section(self, section_name: str) -> Section:
        """Get Mdu Section.

        Returns:
            Section:
                Section object.
                    start (int):
                        The index of the section header line (0-index).
                    end (int):
                        The index of the last line in the section (0-index), the last line in the section is the line before the
                        header of the next section.
                    non_key_value_lines_at_end (int):
                        number of empty or non-key-value lines at the end of the section.
                    last_key_value_line_index (int):
                        index of the last key-value line in the section (0-based index), use this index to append any key-value
                        to the end of the section.
        """
        return Section(section_name, self.content)

    def get_inifield_file(
        self,
        inifield_file: Optional[PathOrStr],
    ) -> Path:
        ini_field_file = self.get_keyword(INIFIELD_FILE_LINE)
        root_dir = self.mdu_path.parent
        if inifield_file is not None:
            # user defined initial field file
            inifield_file = root_dir / inifield_file
        elif isinstance(ini_field_file, Path):
            # from the LegacyFMModel
            inifield_file = ini_field_file.resolve()
        elif isinstance(ini_field_file, str):
            # from reading the geometry section
            inifield_file = root_dir / ini_field_file
        else:
            print(
                f"The initial field file is not found in the mdu file, and not provided by the user. \n "
                f"given: {inifield_file}."
            )
        return inifield_file

    def get_structure_file(
        self,
        usr_structure_file: Optional[PathOrStr],
    ) -> Path:
        structure_file = self.get_keyword(STRUCTURE_FILE_LINE)

        if usr_structure_file is not None:
            # user defined structure file
            usr_structure_file = self.mdu_path / usr_structure_file
        elif isinstance(structure_file, Path):
            # from the LegacyFMModel
            usr_structure_file = structure_file.resolve()
        elif isinstance(structure_file, str):
            # from reading the geometry section
            usr_structure_file = self.mdu_path / structure_file
        else:
            print(
                "The structure file is not found in the mdu file, and not provide by the user. \n"
                f"given: {usr_structure_file}."
            )
        return usr_structure_file


def save_mdu_file(content: List[str], output_path: PathOrStr) -> None:
    """Save the updated MDU file content to disk.

    Args:
        content: The updated content of the MDU file
        output_path: The path where the updated MDU file should be saved
    """
    with open(output_path, "w", encoding="utf-8") as f:
        f.writelines(content)


def get_ref_time(input_date: str, date_format: str = "%Y%m%d"):
    """Convert a date string to a datetime object."""
    date_object = datetime.strptime(f"{input_date}", date_format)
    return f"MINUTES SINCE {date_object}"



from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

from pandas import DataFrame
from pydantic import Field, ValidationInfo, field_validator

from hydrolib.core.base.models import BaseModel, ModelSaveSettings, ParsableFileModel
from hydrolib.core.base.utils import FortranUtils
from hydrolib.core.dflowfm.tim.parser import TimParser
from hydrolib.core.dflowfm.tim.serializer import TimSerializer, TimSerializerConfig


class TimRecord(BaseModel):
    """Single tim record, representing a time and a list of data."""

    time: float
    """float: Time of the time record."""

    data: List[float] = Field(default_factory=list)
    """List[float]: Record of the time record."""


class TimModel(ParsableFileModel):
    """Class representing a tim (*.tim) file.

    Attributes:
        serializer_config (TimSerializerConfig):
            Configuration for serialization of the .tim file.
        comments (List[str]):
            Header comments from the .tim file.
        timeseries (List[TimRecord]):
            A list of TimRecord objects, each containing a time value and associated data.
        quantities_names (Optional[List[str]]):
            List of names for the quantities in the timeseries.


    Methods:
        _ext() -> str:
            Returns the file extension for .tim files.
        _filename() -> str:
            Returns the default filename for .tim files.
        _get_serializer() -> Callable:
            Returns the serializer callable for .tim files.
        _get_parser() -> Callable:
            Returns the parser callable for .tim files.
        _validate_timeseries_values(cls, v: List[TimRecord]) -> List[TimRecord]:
            Validates the timeseries data.
        as_dataframe(columns: List[Any] = None) -> DataFrame:
            Returns the timeseries as a pandas DataFrame.
        _validate_quantities_names(cls, v, values) -> List[str]:
            Validates that quantities_names match the values or each record.

    Raises:
        ValueError:
            If the timeseries has inconsistent column counts or duplicate time values.

    Examples:
        Create a TimModel object from a .tim file:

            ```python
            >>> from hydrolib.core.dflowfm.tim.models import TimModel, TimRecord
            >>> tim_model = TimModel(filepath="tests/data/input/tim/triple_data_for_timeseries.tim")
            >>> print(tim_model.timeseries)
            [TimRecord(time=10.0, data=[1.232, 2.343, 3.454]), TimRecord(time=20.0, data=[4.565, 5.676, 6.787]), TimRecord(time=30.0, data=[1.5, 2.6, 3.7])]

            ```

        Provide names for the quantities in the timeseries:
            ```python
            >>> quantities_names = ["discharge", "waterlevel", "salinity", "temperature", "initialtracer"]
            >>> tim_model = TimModel(filepath="tests/data/input/source-sink/tim-5-columns.tim", quantities_names=quantities_names)
            >>> print(tim_model.quantities_names)
            ['discharge', 'waterlevel', 'salinity', 'temperature', 'initialtracer']
            >>> print(tim_model.as_dataframe())
                   discharge  waterlevel  salinity  temperature  initialtracer
            0.0          1.0         2.0       3.0          4.0            5.0
            100.0        1.0         2.0       3.0          4.0            5.0
            200.0        1.0         2.0       3.0          4.0            5.0
            300.0        1.0         2.0       3.0          4.0            5.0
            400.0        1.0         2.0       3.0          4.0            5.0

            ```

        Create a `TimModel` object from a dictionary:
            ```python
            >>> data = {
            ...     "comments": ["# Example comment"],
            ...     "timeseries": [TimRecord(time=0.0, data=[1.0, 2.0])]
            ... }
            >>> tim_model = TimModel(**data)
            >>> print(tim_model.timeseries)
            [TimRecord(time=0.0, data=[1.0, 2.0])]

            ```

        Create `TimModel` from `TimRecord` objects:
            ```python
            >>> new_tim = TimModel()
            >>> new_tim.comments = ["# Example comment"]
            >>> new_tim.timeseries = [TimRecord(time=0.0, data=[1.0, 2.0])]

            ```

        Serialize the `TimModel` to a .tim file:
            ```python
            >>> new_tim.save(filepath=Path("output.tim")) # doctest: +SKIP

            ```

    See Also:
        TimParser: Used for parsing .tim files.
        TimSerializer: Used for serializing .tim files.
        TimRecord: Represents individual time and data entries in the timeseries.

    Notes:
        This class ensures the integrity of the timeseries by validating data consistency and detecting duplicate time entries.

    References:
        - `TIM file format <https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#C4>`_
    """

    serializer_config: TimSerializerConfig = TimSerializerConfig()
    """TimSerializerConfig: The serialization configuration for the tim file."""

    comments: List[str] = Field(default_factory=list)
    """List[str]: A list with the header comment of the tim file."""

    timeseries: List[TimRecord] = Field(default_factory=list)
    """List[TimRecord]: A list containing the timeseries."""

    quantities_names: Optional[List[str]] = Field(default=None)

    def __init__(
        self,
        filepath: Optional[Union[str, Path]] = None,
        quantities_names: Optional[List[str]] = None,
        **parsable_file_kwargs: Any,
    ):
        """
        Custom initializer to handle extra parameters specific to TimModel.

        Args:
            filepath (Path):
                Path to the .tim file.
            quantities_names (Optional[List[str]]):
                Names for the quantities in the timeseries.
            **parsable_file_kwargs (Any):
                Other arguments for the superclass.
        """
        super().__init__(filepath=filepath, **parsable_file_kwargs)
        self.quantities_names = quantities_names

    @classmethod
    def _ext(cls) -> str:
        return ".tim"

    @classmethod
    def _filename(cls) -> str:
        return "timeseries"

    @classmethod
    def _get_serializer(
        cls,
    ) -> Callable[[Path, Dict, TimSerializerConfig, ModelSaveSettings], None]:
        return TimSerializer.serialize

    @classmethod
    def _get_parser(cls) -> Callable[[Path], Dict]:
        return TimParser.parse

    @field_validator("timeseries", mode="before", check_fields=True)
    def replace_fortran_scientific_notation_for_floats(cls, value, field):
        for record in value:
            if isinstance(record, dict):
                record["time"] = FortranUtils.replace_fortran_scientific_notation(
                    record["time"]
                )
                record["data"] = FortranUtils.replace_fortran_scientific_notation(
                    record["data"]
                )
            elif isinstance(record, TimRecord):
                record.time = FortranUtils.replace_fortran_scientific_notation(
                    record.time
                )
                record.data = FortranUtils.replace_fortran_scientific_notation(
                    record.data
                )

        return value

    @field_validator("timeseries", mode="after")
    @classmethod
    def _validate_timeseries_values(cls, v: List[TimRecord]) -> List[TimRecord]:
        """Validate if the number of columns per timeseries matches and if the timeseries have no duplicate times.

        Args:
            v (List[TimRecord]): Timeseries to validate.

        Raises:
            ValueError: When the number of columns for timeseries is zero.
            ValueError: When the number of columns differs per timeseries.
            ValueError: When the timeseries has a duplicate time.

        Returns:
            List[TimRecord]: Validated timeseries.
        """
        if len(v) == 0:
            return v

        cls._raise_error_if_amount_of_columns_differ(v)
        cls._raise_error_if_duplicate_time(v)

        return v

    @staticmethod
    def _raise_error_if_amount_of_columns_differ(timeseries: List[TimRecord]) -> None:
        n_columns = len(timeseries[0].data)

        if n_columns == 0:
            raise ValueError("Time series cannot be empty.")

        for timrecord in timeseries:
            if len(timrecord.data) != n_columns:
                raise ValueError(
                    f"Time {timrecord.time}: Expected {n_columns} columns, but was {len(timrecord.data)}"
                )

    @staticmethod
    def _raise_error_if_duplicate_time(timeseries: List[TimRecord]) -> None:
        seen_times = set()
        for timrecord in timeseries:
            if timrecord.time in seen_times:
                raise ValueError(
                    f"Timeseries cannot contain duplicate times. Time: {timrecord.time} is duplicate."
                )
            seen_times.add(timrecord.time)

    @field_validator("quantities_names", mode="before")
    def _validate_quantities_names(cls, v, info: ValidationInfo) -> List[str]:
        """Validate if the number of quantities_names matches the number of columns in the timeseries.

        The validator compared the amount of quantities_names with the number of columns in the first record of
        the timeseries.
        """
        if v is not None:
            first_records_data = info.data["timeseries"][0].data
            if len(v) != len(first_records_data):
                raise ValueError(
                    f"The number of quantities_names ({len(v)}) must match the number of columns in the Tim file ({len(first_records_data)})."
                )
        return v

    def add_column(self, new_values: List[float], column_name: str = None) -> None:
        """
        Add new values to each TimRecord in the timeseries, representing a new location.

        Args:
            new_values (List[float]): A list of new values to add, one for each TimRecord.
            column_name (str, optional): The name of the new column. Defaults to None.
                if None, the column is named as "quantity-{len(quantities_names) + 1}".

        Raises:
            ValueError: If the number of new values does not match the number of TimRecords.

        Examples:
            ```python
            >>> tim_model = TimModel(
            ...     timeseries=[
            ...         TimRecord(time=0.0, data=[1.0, 2.0]),
            ...         TimRecord(time=1.0, data=[3.0, 4.0]),
            ...     ]
            ... )
            >>> tim_model.add_column([5.0, 6.0])
            >>> print(tim_model.timeseries)
            [TimRecord(time=0.0, data=[1.0, 2.0, 5.0]), TimRecord(time=1.0, data=[3.0, 4.0, 6.0])]

            ```
        """
        if len(new_values) != len(self.timeseries):
            raise ValueError(
                f"Expected {len(self.timeseries)} values, but got {len(new_values)}."
            )

        for record, value in zip(self.timeseries, new_values):
            record.data.append(value)

        if self.quantities_names:
            if column_name is None:
                column_name = f"quantity-{len(self.quantities_names) + 1}"
            self.quantities_names.append(column_name)

    def as_dataframe(self, columns: List[Any] = None) -> DataFrame:
        """Return the timeseries as a pandas DataFrame.

        Args:
            columns (List[Any, str], optional, Defaults to None):
                The column names for the DataFrame.

        Returns:
            DataFrame: The timeseries as a pandas DataFrame.

        Notes:
            - If the columns are not provided, the quantities_names will be used as column names.
            - If the quantities_names are not provided, the columns will be named as 0, 1, 2, etc.

        Examples:
            Create a `TimModel` object from a .tim file:
                >>> from hydrolib.core.dflowfm.tim.models import TimModel
                >>> tim_model = TimModel(filepath="tests/data/input/tim/triple_data_for_timeseries.tim")
                >>> df = tim_model.as_dataframe()
                >>> print(df)
                          0      1      2
                10.0  1.232  2.343  3.454
                20.0  4.565  5.676  6.787
                30.0  1.500  2.600  3.700

            If the `TimModel` object was created with quantities names:
                >>> quantities_names = ["Column1", "Column2", "Column3"]
                >>> tim_model = TimModel(filepath="tests/data/input/tim/triple_data_for_timeseries.tim", quantities_names=quantities_names)
                >>> df = tim_model.as_dataframe()
                >>> print(df)
                      Column1  Column2  Column3
                10.0    1.232    2.343    3.454
                20.0    4.565    5.676    6.787
                30.0    1.500    2.600    3.700

            To add column names to the DataFrame after you have created the `TimModel` object:
                >>> df = tim_model.as_dataframe(columns=["Column1", "Column2", "Column3"])
                >>> print(df)
                      Column1  Column2  Column3
                10.0    1.232    2.343    3.454
                20.0    4.565    5.676    6.787
                30.0    1.500    2.600    3.700
        """
        time_series = [record.data for record in self.timeseries]
        index = [record.time for record in self.timeseries]
        if not columns:
            columns = self.quantities_names
        return DataFrame(time_series, index=index, columns=columns)

    def as_dict(self) -> Dict[str, List[float]]:
        """Extract time series data from a TIM model.

        Extract the time series data (each column) from the TimModel object

        Returns:
            Dict[str, List[float]]: A dictionary containing the time series data form each column.
            the keys of the dictionary will be index starting from 1 to the number of columns in the tim file
            (excluding the first column(time)).

        Examples:
            ```python
            >>> tim_file = Path("tests/data/input/source-sink/leftsor.tim")
            >>> time_file = TimParser.parse(tim_file)
            >>> tim_model = TimModel(**time_file)
            >>> time_series = tim_model.as_dict()
            >>> print(time_series) # doctest: +SKIP
            {
                1: [1.0, 1.0, 3.0, 5.0, 8.0],
                2: [2.0, 2.0, 5.0, 8.0, 10.0],
                3: [3.0, 5.0, 12.0, 9.0, 23.0],
                4: [4.0, 4.0, 4.0, 4.0, 4.0]
            }
            ```
        """
        data = self.as_dataframe().to_dict(orient="list")
        return data

    def get_units(self) -> Optional[List[str]]:
        """Return the units for each quantity in the timeseries.

        Returns:
            List[str]: A list of units for each quantity in the timeseries.

        Examples:
            Create a `TimModel` object from a .tim file:
                ```python
                >>> from hydrolib.core.dflowfm.tim.models import TimModel
                >>> tim_model = TimModel(filepath="tests/data/input/source-sink/tim-5-columns.tim")
                >>> tim_model.quantities_names = ["discharge", "waterlevel", "temperature", "salinity", "initialtracer"]
                >>> print(tim_model.get_units())
                ['m3/s', 'm', 'degC', '1e-3', '-']

                ```
        """
        if self.quantities_names is None:
            return None
        return TimModel._get_quantity_unit(self.quantities_names)



from pathlib import Path
from typing import Dict, List, Tuple


class BaseParser:
    @staticmethod
    def _read_header_comments(lines: List[str]) -> Tuple[List[str], int]:
        """Read the header comments of the lines from the file.
        The comments are only expected at the start of the file.
        When a non comment line is encountered, all comments from the header will be retuned together with the start index of the data.

        Args:
            lines (List[str]): Lines from the the file which is read.

        Returns:
            Tuple of List[str] and int, the List[str] contains the commenst from the header, the int is the start index of the data.
        """
        comments: List[str] = []
        start_timeseries_index = 0
        for line_index, line in enumerate(lines):

            line = line.strip()

            if len(line) == 0:
                comments.append(line)
                continue

            if line.startswith("#") or line.startswith("*"):
                comments.append(line[1:])
                continue

            start_timeseries_index = line_index
            break

        return comments, start_timeseries_index

    @staticmethod
    def _raise_error_if_contains_comment(line: str, line_index: int) -> None:
        if "#" in line or "*" in line:
            raise ValueError(
                f"Line {line_index}: comments are only supported at the start of the file, before the data."
            )


class DummmyParser:
    @staticmethod
    def parse(filepath: Path) -> Dict:
        return {}



from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    """Configuration management, can be derived from ENV of .env file."""

    model_config = SettingsConfigDict(
        env_file=".env",  # read from .env if present
        env_file_encoding="utf-8",  # typical default
        extra="ignore",  # ignore unknown env vars
    )

    FM_EXECUTABLE: str = "fm.exe"


settings = Settings()
